#! /usr/bin/env python

import numpy

from landlab.grid.base import ModelGrid
import landlab.utils.structured_grid as sgrid
from landlab.utils import count_repeated_values
from landlab.grid.base import (INTERIOR_NODE, FIXED_VALUE_BOUNDARY,
                               FIXED_GRADIENT_BOUNDARY, TRACKS_CELL_BOUNDARY,
                               INACTIVE_BOUNDARY, BAD_INDEX_VALUE,
                              )


def node_has_boundary_neighbor(mg, id):
    for neighbor in mg.get_neighbor_list(id):
        if mg.node_status[neighbor] != INTERIOR_NODE:
            return True
    for neighbor in mg.get_diagonal_list(id):
        if mg.node_status[neighbor] != INTERIOR_NODE:
            return True
    return False


has_boundary_neighbor = numpy.vectorize(node_has_boundary_neighbor,
                                        excluded=['mg'])


class RasterModelGrid(ModelGrid):
    """
    This inherited class implements a regular, raster 2D grid with uniform
    cell dimensions.
    
    Examples:
        
        >>> rmg = RasterModelGrid()
        >>> rmg.num_nodes
        0
        >>> rmg = RasterModelGrid(4, 5, 1.0) # rows, columns, spacing
        >>> rmg.num_nodes
        20
    """

    def __init__(self, num_rows=0, num_cols=0, dx=1.0):
        """
        Optionally takes numbers of rows and columns and cell size as
        inputs. If this are given, calls initialize() to set up the grid.
        
        """
        #print 'RasterModelGrid.init'
        
        # Set number of nodes, and initialize if caller has given dimensions
        self.ncells = num_rows * num_cols   #TBX
        self.num_nodes = num_rows * num_cols
        if self.num_nodes > 0:
            self.initialize( num_rows, num_cols, dx )

    def initialize( self, num_rows, num_cols, dx ):
        """
        Sets up a num_rows by num_cols grid with cell spacing dx and
        (by default) regular boundaries (that is, all perimeter cells are
        boundaries and all interior cells are active).

        To be consistent with unstructured grids, the raster grid is
        managed not as a 2D array but rather as a set of vectors that
        describe connectivity information between nodes, links, active links,
        cells, active cells, faces, patches, junctions, and corners.
        
        Examples and doctests:

            >>> import landlab as ll
            >>> rmg = RasterModelGrid()
            >>> numrows = 20          # number of rows in the grid
            >>> numcols = 30          # number of columns in the grid
            >>> dx = 10.0             # grid cell spacing
            >>> rmg.initialize(numrows, numcols, dx)
            >>> rmg.num_nodes,rmg.num_cells,rmg.num_links,rmg.num_active_links
            (600, 504, 1150, 1054)
            >>> rmg = RasterModelGrid(4, 5)
            >>> rmg.num_nodes,rmg.num_cells,rmg.num_links,rmg.num_active_links
            (20, 6, 31, 17)
            >>> rmg.node_status
            array([1, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1], dtype=int8)
            >>> rmg.node_activecell[3] == ll.BAD_INDEX_VALUE
            True
            >>> rmg.node_activecell[8]
            2
            >>> rmg.node_numinlink
            array([0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 1, 2, 2, 2, 2, 1, 2, 2, 2, 2])
            >>> rmg.node_inlink_matrix
            array([[-1, 15, 16, 17, 18,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11,
                    12, 13, 14],
                   [-1, -1, -1, -1, -1, -1, 19, 20, 21, 22, -1, 23, 24, 25, 26, -1, 27,
                    28, 29, 30]])
            >>> rmg.node_numoutlink
            array([2, 2, 2, 2, 1, 2, 2, 2, 2, 1, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0])
            >>> rmg.node_outlink_matrix[0]
            array([ 0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 27, 28,
                   29, 30, -1])
            >>> rmg.node_numactiveinlink
            array([0, 0, 0, 0, 0, 0, 2, 2, 2, 1, 0, 2, 2, 2, 1, 0, 1, 1, 1, 0])
            >>> rmg.node_active_inlink_matrix
            array([[-1, -1, -1, -1, -1, -1,  0,  1,  2, 12, -1,  3,  4,  5, 16, -1,  6,
                     7,  8, -1],
                   [-1, -1, -1, -1, -1, -1,  9, 10, 11, -1, -1, 13, 14, 15, -1, -1, -1,
                    -1, -1, -1]])
            >>> rmg.node_numactiveoutlink
            array([0, 1, 1, 1, 0, 1, 2, 2, 2, 0, 1, 2, 2, 2, 0, 0, 0, 0, 0, 0])
            >>> rmg.node_active_outlink_matrix
            array([[-1,  0,  1,  2, -1,  9,  3,  4,  5, -1, 13,  6,  7,  8, -1, -1, -1,
                    -1, -1, -1],
                   [-1, -1, -1, -1, -1, -1, 10, 11, 12, -1, -1, 14, 15, 16, -1, -1, -1,
                    -1, -1, -1]])
            >>> rmg.cell_node
            array([ 6,  7,  8, 11, 12, 13])
            >>> rmg.link_fromnode
            array([ 0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14,  0,  1,
                    2,  3,  5,  6,  7,  8, 10, 11, 12, 13, 15, 16, 17, 18])
            >>> rmg.link_tonode
            array([ 5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19,  1,  2,
                    3,  4,  6,  7,  8,  9, 11, 12, 13, 14, 16, 17, 18, 19])
            >>> rmg.link_face[20]
            10
            >>> rmg.active_links
            array([ 1,  2,  3,  6,  7,  8, 11, 12, 13, 19, 20, 21, 22, 23, 24, 25, 26])
        """
        
        if self.DEBUG_TRACK_METHODS:
            print 'RasterModelGrid.initialize('+str(num_rows)+', ' \
                   +str(num_cols)+', '+str(dx)+')'
        
        # Basic info about raster size and shape
        self.nrows = num_rows
        self.ncols = num_cols
        self.ncells = num_rows * num_cols # soon to be deprecated, June 2013
        self._dx = dx
        self.cellarea = dx*dx
        self.num_nodes = num_rows * num_cols
        self.num_cells = (num_rows-2) * (num_cols-2)
        self.num_active_cells = self.num_cells
        self.num_links = num_cols*(num_rows-1)+num_rows*(num_cols-1)
        self.num_active_links = self.num_links-(2*(num_cols-1)+2*(num_rows-1))
        
        # We need at least one row or column of boundary cells on each
        # side, so the grid has to be at least 3x3
        assert(numpy.min((num_rows, num_cols)) >= 3)

        # Record number of boundary and interior cells and the number
        # of interior faces. Ultimately, this info could be overridden
        # if using an irregular geometry of "interior" cells within the
        # rectangular domain. Note that we don't include any faces
        # between boundary cells.
        # NG Do we still have boundary cells?  I thought that cells were only
        # defined on interior nodes.
        self.n_boundary_cells = 2 * ( num_rows - 2 ) + 2 * ( num_cols - 2 ) + 4
        self.n_interior_cells = self.ncells - self.n_boundary_cells
        self.num_faces = ( num_rows - 1 ) * ( num_cols - 2 ) + \
                      ( num_rows - 2 ) * ( num_cols - 1 )
        self.nfaces = self.num_faces # TBX; for backward compatibility
        if self.DEBUG_VERBOSE:
            print self.num_faces
        
        # Assign and store node (x,y,z) coordinates.
        #
        # The relation between node (x,y) coordinates and position is
        # illustrated here for a five-column, four-row grid. The numbers show
        # node positions, and the - and | symbols show the links connecting
        # the nodes.
        #
        # 15------16------17------18------19
        #  |       |       |       |       |
        #  |       |       |       |       |
        #  |       |       |       |       |
        # 10------11------12------13------14
        #  |       |       |       |       |
        #  |       |       |       |       |   
        #  |       |       |       |       |
        #  5-------6-------7-------8-------9
        #  |       |       |       |       |
        #  |       |       |       |       |
        #  |       |       |       |       |
        #  0-------1-------2-------3-------4
        #
        (self._node_x, self._node_y) = sgrid.node_coords(
            (num_rows, num_cols), (self._dx, self._dx), (0., 0.))

        # Node boundary/active status:
        # Next, we set up an array of "node status" values, which indicate 
        # whether a given node is an active, non-boundary node, or some type of 
        # boundary. Here we default to having all perimeter nodes be active
        # fixed-value boundaries.
        self.node_status = sgrid.node_status(
            self.shape, boundary_status=FIXED_VALUE_BOUNDARY)
        
        # Cell lists:
        # For all cells, we create a list of the corresponding node ID for 
        # each cell.
        # We also have a list of the cell IDs of all active cells. By default,
        # all cells are active, so for example if there are six cells, the
        # self.active_cells list reads: 0, 1, 2, 3, 4, 5
        # 
        # Cells and faces in a five-column, four-row grid look like this
        # (where the numbers are cell IDs and lines show faces):
        #
        # |-------|-------|-------|
        # |       |       |       |
        # |   3   |   4   |   5   |
        # |       |       |       |
        # |-------|-------|-------|
        # |       |       |       |
        # |   0   |   1   |   2   |
        # |       |       |       |
        # |-------|-------|-------|
        #
        # While we're at it, we will also build the node_activecell list. This
        # list records, for each node, the ID of its associated active cell, 
        # or None if it has no associated active cell (i.e., it is a boundary)
        self.cell_node = sgrid.cell_node_index((num_rows, num_cols))
        self.node_activecell = sgrid.node_active_cell((num_rows, num_cols))
        self.active_cells = sgrid.active_cells((num_rows, num_cols))
        self.activecell_node = self.cell_node.copy()

        # Link lists:
        # For all links, we encode the "from" and "to" nodes, and the face
        # (if any) associated with the link. If the link does not intersect a
        # face, then face is assigned None.
        # For active links, we store the corresponding link ID.
        #
        # The numbering scheme for links in RasterModelGrid is illustrated with
        # the example of a five-column by four-row grid (each * is a node,
        # the lines show links, and the ^ and > symbols indicate the direction
        # of each link: up for vertical links, and right for horizontal ones):
        #
        #  *--27-->*--28-->*--29-->*--30-->*
        #  ^       ^       ^       ^       ^
        # 10      11      12      13      14
        #  |       |       |       |       |
        #  *--23-->*--24-->*--25-->*--26-->*
        #  ^       ^       ^       ^       ^
        #  5       6       7       8       9   
        #  |       |       |       |       |
        #  *--19-->*--20-->*--21-->*--22-->*
        #  ^       ^       ^       ^       ^
        #  0       1       2       3       4
        #  |       |       |       |       |
        #  *--15-->*--16-->*--17-->*--18-->*
        #
        #   create the fromnode and tonode lists
        (self.link_fromnode,
         self.link_tonode) = sgrid.node_link_index((num_rows, num_cols))

        #   set up in-link and out-link matrices and numbers
        self.setup_inlink_and_outlink_matrices()
        
        #   set up the list of active links
        self.reset_list_of_active_links()

        #   set up link faces
        #
        #   Here we assume that we've already created a list of active links
        # in which all 4 boundaries are "open", such that each boundary node
        # (except the 4 corners) is connected to an adjacent interior node. In
        # this case, there will be the same number of faces as active links,
        # and the numbering of faces will be the same as the corresponding
        # active links. We start off creating a list of all None values. Only
        # those links that cross a face will have this None value replaced with
        # a face ID.
        self.link_face = sgrid.link_faces((num_rows, num_cols),
                                          actives=self.active_links)

        # List of neighbors for each cell: we will start off with no
        # list. If a caller requests it via get_neighbor_list or
        # create_neighbor_list, we'll create it if necessary.
        self.neighbor_list_created = False

        # List of diagonal neighbors. As with the neighbor list, we'll only
        # create it if requested.
        self.diagonal_list_created = False

    def _setup_cell_areas_array(self):
        self.active_cell_areas = numpy.empty(self.num_active_cells)
        self.active_cell_areas.fill(self._dx ** 2)
        return self.active_cell_areas

    @property
    def shape(self):
        return (self.nrows, self.ncols)

    @property
    def dx(self):
        return self._dx

    @property
    def shape(self):
        return (self.nrows, self.ncols)

    def setup_inlink_and_outlink_matrices(self):
        """
        Creates data structures to record the numbers of inlinks and outlinks
        for each node. An inlink of a node is simply a link that has the node as
        its "to" node, and an outlink is a link that has the node as its "from".
        
        We store the inlinks in a 2-row by num_nodes-column matrix called
        node_inlink_matrix. It has two rows because we know that the nodes in
        our raster grid will never have more than two inlinks an two outlinks
        each (a given node could also have zero or one of either). The outlinks
        are stored in a similar matrix.
        
        We also keep track of the total number of inlinks and outlinks at each
        node in the num_inlinks and num_outlinks arrays.
        
        The inlink and outlink matrices are useful in numerical calculations.
        Each row of each matrix contains one inlink or outlink per node. So, if
        you have a corresponding "flux" matrix, you can map incoming or
        outgoing fluxes onto the appropriate nodes. More information on this is
        in the various calculate_flux_divergence... functions.
        
        What happens if a given node does not have two inlinks or outlinks? We
        simply put the default value -1 in this case. This allows us to use a 
        cute little trick when computing inflows and outflows. We make our 
        "flux" array one element longer than the number of links, with the last
        element containing the value 0. Thus, any time we add an influx from 
        link number -1, Python takes the value of the last element in the array,
        which is zero. By doing it this way, we maintain the efficiency that 
        comes with the use of numpy. Again, more info can be found in the 
        description of the flux divergence functions.
        
        Example:
            
            >>> rmg = RasterModelGrid(4, 5, 1.0)
        """

        (self.node_inlink_matrix,
         self.node_numinlink) = sgrid.setup_inlink_matrix(
             self.shape, tonodes=self.link_tonode)

        (self.node_outlink_matrix,
         self.node_numoutlink) = sgrid.setup_outlink_matrix(
             self.shape, fromnodes=self.link_fromnode)
        
    def setup_active_inlink_and_outlink_matrices(self):
        """
        Creates data structures to record the numbers of active inlinks and 
        active outlinks for each node. These data structures are equivalent to
        the "regular" inlink and outlink matrices, except that it uses the IDs
        of active links (only).
        """
        # Create active in-link and out-link matrices.
        self.node_active_inlink_matrix = - numpy.ones((2, self.num_nodes),
                                                       dtype=numpy.int)
        self.node_active_outlink_matrix = - numpy.ones((2, self.num_nodes),
                                                        dtype=numpy.int)
        # Set up the inlink arrays
        tonodes = self.activelink_tonode
        self.node_numactiveinlink = numpy.bincount(tonodes,
                                                   minlength=self.num_nodes)

        counts = count_repeated_values(self.activelink_tonode)
        for (count, (tonodes, active_link_ids)) in enumerate(counts):
            self.node_active_inlink_matrix[count][tonodes] = active_link_ids

        # Set up the outlink arrays
        fromnodes = self.activelink_fromnode
        self.node_numactiveoutlink = numpy.bincount(fromnodes,
                                                    minlength=self.num_nodes)
        counts = count_repeated_values(self.activelink_fromnode)
        for (count, (fromnodes, active_link_ids)) in enumerate(counts):
            self.node_active_outlink_matrix[count][fromnodes] = active_link_ids

    def cell_faces(self, cell_id):
        """
        Returns an array of the face IDs for the faces of a cell with ID,
        *cell_id*. The faces are listed clockwise, starting with the bottom
        face. *cell_id* can be either a scalar or an array. If an array,
        return the faces for each cell of the array.
        
        >>> mg = RasterModelGrid(4, 5)
        >>> mg.cell_faces(0)
        array([ 0,  9,  3, 10])

        >>> mg.cell_faces([0, 5])
        array([[ 0,  9,  3, 10],
               [ 5, 15,  8, 16]])
        """
        node_id = self.cell_node[cell_id]
        inlinks = self.node_inlink_matrix[:, node_id].T
        outlinks = self.node_outlink_matrix[:, node_id].T
        return numpy.concatenate(
            (self.link_face[inlinks], self.link_face[outlinks]), axis=1)

    def get_grid_xdimension(self):
        '''
        Returns the x dimension of the grid. Method added 5/1/13 by DEJH.
        '''
        return (self.ncols * self._dx)
    
    def get_grid_ydimension(self):
        '''
        Returns the y dimension of the grid. Method added 5/1/13 by DEJH.
        '''
        return (self.nrows * self._dx)
        
    def get_count_of_interior_nodes(self):
        """
        Returns the number of interior nodes on the grid.
        """
        return sgrid.interior_node_count(self.shape)
        
    def get_count_of_all_nodes(self):
        """
        Returns total number of nodes, including boundaries.
        """
        return(self.num_nodes)
    
    def get_count_of_cols(self):
        """
        Returns the number of columns, including boundaries.  
        NG, June 2013
        """
        return(self.ncols)
        
    def get_count_of_rows(self):
        """
        Returns the number of rows, including boundaries.  
        NG, June 2013
        """
        return(self.nrows)
    
    def get_grid_spacing(self):
        """
        Returns the spacing between grid nodes.
        DEJH July 2013
        """
        return(self._dx)

    def get_nodes_around_point(self, xcoord, ycoord):
        """
        This method takes an x,y coordinate within the grid, then returns the 
        IDs of the four nodes of the area (enclosure?) around that point as a 
        4 item list. Because the geometry of this grid is so simple, it works 
        purely by counting the number of squares left and below the point. 
        Method added 4/29/13 by DEJH.
        """
        ID = int(ycoord//self._dx * self.ncols + xcoord//self._dx)
        return numpy.array([ID, ID+self.ncols, ID+self.ncols+1, ID+1])
    
    def get_minimum_active_link_length(self):
        """
        Returns the horizontal length of the shortest active link in the grid.
        Overrides ModelGrid.get_minimum_active_link_length().
        """
        return self._dx

    def calculate_max_gradient_across_node(self, u, cell_id):
        '''
            This method calculates the gradients in u across all 4 faces of the 
            cell with ID cell_id, and across the four diagonals. It then returns 
            the steepest (most negative) of these values, followed by its dip 
            direction (e.g.: 0.12, 225). i.e., this is a D8 algorithm. Slopes 
            downward from the cell are reported as positive.
            
            This code is actually calculating slopes, not gradients.  
            The max gradient is the most negative, but the max slope is the most
            positive.  So, this was updated to return the max value, not the 
            min.
            
        GT: Might be possible to speed this up using inlink_matrix and 
        outlink_matrix.
        '''
        #We have poor functionality if these are edge cells! Needs an exception
        neighbor_cells = self.get_neighbor_list(cell_id)
        neighbor_cells.sort()
        #print 'Node is internal: ', self.is_interior(cell_id)
        #print 'Neighbor cells: ', neighbor_cells
        diagonal_cells = []
        if neighbor_cells[0]!=-1:
            diagonal_cells.extend([neighbor_cells[0]-1, neighbor_cells[0]+1])
        if neighbor_cells[3]!=-1:
            diagonal_cells.extend([neighbor_cells[3]-1, neighbor_cells[3]+1])
        slopes = []
        diagonal_dx = numpy.sqrt(2.)
        for a in neighbor_cells:
            #ng I think this is actually slope as defined by a geomorphologist,
            #that is -dz/dx and not the gradient (dz/dx)
            single_slope = (u[cell_id] - u[a])/self._dx
            #print 'cell id: ', cell_id
            #print 'neighbor id: ', a
            #print 'cell, neighbor are internal: ', self.is_interior(cell_id), self.is_interior(a)
            #print 'cell elev: ', u[cell_id]
            #print 'neighbor elev: ', u[a]
            #print single_slope
            if not numpy.isnan(single_slope): #This should no longer be necessary, but retained in case
                slopes.append(single_slope)
            else:
                print 'NaNs present in the grid!'
        for a in diagonal_cells:
            single_slope = (u[cell_id] - u[a])/diagonal_dx
            #print single_slope
            if not numpy.isnan(single_slope):
                slopes.append(single_slope)
            else:
                print 'NaNs present in the grid!'
        #print 'Slopes list: ', slopes
        #ng thinks that the maximum slope should be found here, not the 
        #minimum slope, old code commented out.  New code below it.
        #if slopes:
        #    min_slope, index_min = min((min_slope, index_min) for (index_min, min_slope) in enumerate(slopes))
        #else:
        #    print u
        #    print 'Returning NaN angle and direction...'
        #    min_slope = numpy.nan
        #    index_min = 8
        if slopes:
            max_slope, index_max = max((max_slope, index_max) for (index_max, max_slope) in enumerate(slopes))
        else:
            print u
            print 'Returning NaN angle and direction...'
            max_slope = numpy.nan
            index_max = 8
            
        angles = [180., 270., 90., 0., 225., 135., 315., 45., numpy.nan] #This is inefficient
        
        #ng commented out old code
        #return min_slope, angles[index_min]
        return max_slope, angles[index_max]
        
    def find_node_in_direction_of_max_slope(self, u, node_id):
        '''
            This method calculates the slopes (-dz/dx) in u across all 4 faces of 
            the cell with ID node_id, and across the four diagonals. 
            It then returns the node ID in the direction of the steepest 
            (most positive) of these values,  i.e., this is a 
            D8 algorithm. Slopes downward from the cell are reported as positive.
            Based on code from DH, modified by NG, 6/2013
            
            This doesn't deal with the fixed gradient boundary condition.  
            NG is still confused about that one.
            
            NMG Update.  This is super clumsy. 
            
            DEJH update: Gets confused for the lowest node if w/i grid
            (i.e., closed)- will return a higher neighbour, when it should
            return itself. ->  Now returns itself, but a flagging mechanism
            may be preferable. (i.e., unthinking application will result in a
            doubling of discharge at that node.
        '''
        #We have poor functionality if these are closed boundary nodes! 
        neighbor_nodes = self.get_neighbor_list(node_id)
        neighbor_nodes.sort()
        #print 'Node is internal: ', self.is_interior(cell_id)
        #print 'Neighbor cells: ', neighbor_cells
        diagonal_nodes = []
        #NG also think that this won't happen if you are always sending this 
        #function an id of an interior node.  But maybe there is a case where 
        #this would happen?
        if neighbor_nodes[0]!=-1:
            diagonal_nodes.extend([neighbor_nodes[0]-1, neighbor_nodes[0]+1])
        #ng, if neighbor_nodes is sorted, how could [3] be -1?
        #try commenting out.
        #if neighbor_cells[3]!=-1:
        diagonal_nodes.extend([neighbor_nodes[3]-1, neighbor_nodes[3]+1])
        slopes = []
        diagonal_dx = numpy.sqrt(2.)
        for a in neighbor_nodes:
            if self.node_status[a] != INACTIVE_BOUNDARY:
                single_slope = (u[node_id] - u[a])/self._dx
            else:
                single_slope = -9999
            #print 'cell id: ', cell_id
            #print 'neighbor id: ', a
            #print 'status: ', self.node_status[a]
            #print 'cell, neighbor are internal: ', self.is_interior(cell_id), self.is_interior(a)
            #print 'cell elev: ', u[cell_id]
            #print 'neighbor elev: ', u[a]
            #print single_slope
            if not numpy.isnan(single_slope): #This should no longer be necessary, but retained in case
                slopes.append(single_slope)
            else:
                print 'NaNs present in the grid!'
        for a in diagonal_nodes:
            if self.node_status[a] != INACTIVE_BOUNDARY:
                single_slope = (u[node_id] - u[a])/diagonal_dx
            else:
                single_slope = -9999
            #print single_slope
            if not numpy.isnan(single_slope):
                slopes.append(single_slope)
            else:
                print 'NaNs present in the grid!'
        #print 'Slopes list: ', slopes
        if slopes:
            max_slope, index_max = max((max_slope, index_max) for (index_max, max_slope) in enumerate(slopes))
        else:
            print u
            print 'Returning NaN angle and direction...'
            max_slope = numpy.nan
            index_max = 8
        
        all_neighbor_nodes=numpy.concatenate((neighbor_nodes,diagonal_nodes))
        #print 'all_neighbor_cells ', all_neighbor_cells
        
        #Final check to  allow correct handling of internally draining nodes; DEJH Aug 2013.
        #This remains extremely ad-hoc. An internal node points to itself, but this should never
        #be used to actually route flow. In flow_accumulation, there is an explicit check that flow
        #is not routed to yourself.
        steepest_node = all_neighbor_nodes[index_max]
        #...now if a node is the lowest thing, this method returns that node, not a neighbor:
        if u[steepest_node] > u[node_id]:
            steepest_node=node_id
            #When accumulating flow, it is ESSENTIAL to check if a node routes to itself!!
            #An alternative, creating a list of the internally draining cells
            #try:
            #    self.internal_draining_nodes.append(node_id)
            #except:
            #    self.internal_draining_nodes = [node_id]
            #self.node_status[node_id] = INTERNAL_DRAINING_NODE #==5? ...this doesn't exist yet.
        
        return steepest_node
        
    def set_inactive_boundaries(self, bottom_is_inactive, right_is_inactive, 
                                top_is_inactive, left_is_inactive):
        """
        Handles boundary conditions by setting each of the four sides of the 
        rectangular grid to either 'inactive' or 'active (fixed value)' status.
        Arguments are booleans indicating whether the bottom, right, top, and
        left are inactive (True) or not (False).
        
        For an inactive boundary:
            - the nodes are flagged INACTIVE_BOUNDARY
            - the links between them and the adjacent interior nodes are
              inactive (so they appear on link-based lists, but not
              active_link-based lists)
              
        This means that if you call the calculate_gradients_at_active_links
        method, the inactive boundaries will be ignored: there can be no
        gradients or fluxes calculated, because the links that connect to that
        edge of the grid are not included in the calculation. So, setting a
        grid edge to INACTIVE_BOUNDARY is a convenient way to impose a no-flux
        boundary condition. Note, however, that this applies to the grid as a
        whole, rather than a particular variable that you might use in your
        application. In other words, if you want a no-flux boundary in one
        variable but a different boundary condition for another, then use 
        another method.
        
        The following example sets the top and left boundaries as inactive in a
        four-row by five-column grid that initially has all boundaries active
        and all boundary nodes coded as FIXED_VALUE_BOUNDARY (=1):
        
        >>> rmg = RasterModelGrid(4, 5, 1.0) # rows, columns, spacing
        >>> rmg.num_active_links
        17
        >>> rmg.node_status
        array([1, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1], dtype=int8)
        >>> rmg.set_inactive_boundaries(False, False, True, True)
        >>> rmg.num_active_links
        12
        >>> rmg.node_status
        array([1, 1, 1, 1, 1, 4, 0, 0, 0, 1, 4, 0, 0, 0, 1, 4, 4, 4, 4, 4], dtype=int8)
        
        Note that the four corners are treated as follows:
            bottom left = BOTTOM
            bottom right = RIGHT
            top right = TOP
            top left = LEFT
        """
        if self.DEBUG_TRACK_METHODS:
            print 'ModelGrid.set_inactive_boundaries'
            
        bottom_edge = range(0,self.ncols-1)
        right_edge = range(self.ncols-1,self.num_nodes-1,self.ncols)
        top_edge = range((self.nrows-1)*self.ncols+1,self.num_nodes)
        left_edge = range(self.ncols,self.num_nodes,self.ncols)
            
        if bottom_is_inactive:
            self.node_status[bottom_edge] = INACTIVE_BOUNDARY
        else:
            self.node_status[bottom_edge] = FIXED_VALUE_BOUNDARY

        if right_is_inactive:
            self.node_status[right_edge] = INACTIVE_BOUNDARY
        else:
            self.node_status[right_edge] = FIXED_VALUE_BOUNDARY
            
        if top_is_inactive:
            self.node_status[top_edge] = INACTIVE_BOUNDARY
        else:
            self.node_status[top_edge] = FIXED_VALUE_BOUNDARY

        if left_is_inactive:
            self.node_status[left_edge] = INACTIVE_BOUNDARY
        else:
            self.node_status[left_edge] = FIXED_VALUE_BOUNDARY
        
        self.reset_list_of_active_links()

                
    def set_noflux_boundaries( self, bottom, right, top, left,
                               bc = None ):
        """
        Assigns "no flux" status to one or more sides of the rectangular
        domain, for BoundaryCondition "bc", which default's to ModelGrid's
        self.default_bc (i.e., the default BoundaryCondition used when
        the user doesn't specify another one).

        Boundary cells are either "fixed value" (Dirichlet), which is
        the default, "fixed gradient" (Neumann with a zero derivative), or
        "tracks cell" (they track a cell in the interior on the opposite
        side of the grid, e.g., for periodic). Here we implement no flux
        by mirroring adjacent cells in the interior.

        Boundary status is recorded in the bc's TRACKS_CELL vector, 
        defined in bc's initialize() and N_BOUNDARY_CELLS long. For
        no-flux cells, this vector contains the CID of the neighboring
        cell whose value it will mirror in order to maintain a zero
        gradient. For periodic boundary cells, this vector contains the
        CID of the cell on the opposite side whose value it will track.
        Fixed value cells are indicated by a -1 for TRACKS_CELL.

        For no-flux boundaries, the corner cells (which don't really
        matter much anyway), the neighbor is arbitrarily chosen as either 
        the righthand or lefthand cell.
        
        GT: I believe this is now obsolete, because we can do no-flux simply by 
        setting closed boundaries (Aug 2013).
        """
        
        if bc==None:
            bc = self.default_bc
        
        # For no-flux boundaries, we need to know which interior
        # cells to mirror.
        #self.boundary_nbrs = zeros( self.n_boundary_cells, dtype=numpy.int )
        lower_left = 0
        lower_right = self.ncols-1
        upper_right = self.ncols+self.nrows-2
        upper_left = 2*self.ncols+self.nrows-3

        if bottom:
            for id in xrange( 1, self.ncols-1 ):   # Bottom
                bc.boundary_code[id] = bc.TRACKS_CELL_BOUNDARY
                bc.tracks_cell[id] = id+self.ncols
            bc.boundary_code[lower_left] = bc.TRACKS_CELL_BOUNDARY
            bc.tracks_cell[lower_left] = 1
            bc.boundary_code[lower_right] = bc.TRACKS_CELL_BOUNDARY
            bc.tracks_cell[lower_right] = lower_right-1
        if right:
            nbr = 2*self.ncols-2
            for id in xrange( lower_right+1, upper_right ):   # Right
                bc.boundary_code[id] = bc.TRACKS_CELL_BOUNDARY
                bc.tracks_cell[id] = nbr
                nbr += self.ncols
        if top:
            nbr = self.ncells - (self.ncols+2)
            for id in xrange( upper_right+1, upper_left ):   # Top
                bc.boundary_code[id] = bc.TRACKS_CELL_BOUNDARY
                bc.tracks_cell[id] = nbr
                nbr = nbr - 1
            bc.boundary_code[upper_right] = bc.TRACKS_CELL_BOUNDARY
            bc.tracks_cell[upper_right] = self.ncells-2
            bc.boundary_code[upper_left] = bc.TRACKS_CELL_BOUNDARY
            bc.tracks_cell[upper_left] = self.ncells+1-self.ncols
        if left:
            nbr = (self.nrows-2)*self.ncols + 1
            for id in xrange( upper_left+1, self.n_boundary_cells ):   # Left
                bc.boundary_code[id] = bc.TRACKS_CELL_BOUNDARY
                bc.tracks_cell[id] = nbr
                nbr = nbr - self.ncols
        
        if self.DEBUG_VERBOSE:
            print 'tracks_cell:',bc.tracks_cell
    
    def calculate_gradients_at_active_links(self, s, gradient=None):
        """
        Calculates the gradient in quantity s at each active link in the grid.
        This is nearly identical to the method of the same name in ModelGrid,
        except that it uses self._dx for link length to improve efficiency.
        
        Example:
        
            >>> rmg = RasterModelGrid(4, 5, 1.0)
            >>> u = [0., 1., 2., 3., 0.,
            ...     1., 2., 3., 2., 3.,
            ...     0., 1., 2., 1., 2.,
            ...     0., 0., 2., 2., 0.]
            >>> u = numpy.array(u)
            >>> u
            array([ 0.,  1.,  2.,  3.,  0.,  1.,  2.,  3.,  2.,  3.,  0.,  1.,  2.,
                    1.,  2.,  0.,  0.,  2.,  2.,  0.])
            >>> grad = rmg.calculate_gradients_at_active_links(u)
            >>> grad
            array([ 1.,  1., -1., -1., -1., -1., -1.,  0.,  1.,  1.,  1., -1.,  1.,
                    1.,  1., -1.,  1.])
            
        For greater speed, sending a pre-created numpy array as an argument
        avoids having to create a new one with each call:
            
            >>> grad = numpy.zeros(rmg.num_active_links)
            >>> u = u*10
            >>> grad = rmg.calculate_gradients_at_active_links(u, grad)
            >>> grad
            array([ 10.,  10., -10., -10., -10., -10., -10.,   0.,  10.,  10.,  10.,
                   -10.,  10.,  10.,  10., -10.,  10.])
        """
        if self.DEBUG_TRACK_METHODS:
            print 'RasterModelGrid.calculate_gradients_at_active_links'

        if gradient==None:
            gradient = numpy.zeros(self.num_active_links)
            
        assert (len(gradient)==self.num_active_links), \
                "len(gradient)!=num_active_links"
     
        gradient = (s[self.activelink_tonode]-s[self.activelink_fromnode])/self._dx
        
        return gradient
        
    def calculate_flux_divergence_at_active_cells(self, active_link_flux, 
                                                  net_unit_flux=False):
        """
        Given an array of fluxes along links, computes the net total flux
        within each cell, divides by cell area, and stores the result in
        net_unit_flux. Overrides method of the same name in ModelGrid (nearly
        identical, but uses scalars dx and cellarea instead of variable
        link length and active cell area, respectively).
        
        The function works by calling calculate_flux_divergence_at_nodes, then
        slicing out only the values at active cells. Therefore, it is slower
        than calculate_flux_divergence_at_nodes, even though it returns a
        shorter list of numbers.
        
        The input active_link_flux should be flux of
        something (e.g., mass, momentum, energy) per unit face width, positive
        if flowing in the same direction as its link, and negative otherwise.
        There should be one value per active link. Returns an array of net
        total flux per unit area, one value per active cell (creates this
        array if it is not given as an argument).
          By convention, divergence is positive for net outflow, and negative 
        for net outflow. That's why we *add* outgoing flux and *subtract* 
        incoming flux. This makes net_unit_flux have the same sign and 
        dimensions as a typical divergence term in a conservation equation.

        In general, for a polygonal cell with $N$ sides of lengths
        Li and with surface area A, the net influx divided by cell
        area would be:
            .. math::
                {Q_{net} \over A} = {1 \over A} \sum{q_i L_i}

        For a square cell, which is what we have in RasterModelGrid,
        the sum is over 4 sides of length dx, and
        :math:`A = dx^2`, so:
            .. math::
                {Q_{net} \over A} = {1 \over dx} \sum{q_i}

        .. note::
            The net flux is defined as positive outward, negative
            inward. In a diffusion problem, for example, one would use:
                .. math::
                    {du \over dt} = \\text{source} - \\text{fd}
            where fd is "flux divergence".
            
        Example:
            
            >>> rmg = RasterModelGrid(4, 5, 1.0)
            >>> u = [0., 1., 2., 3., 0.,
            ...      1., 2., 3., 2., 3.,
            ...      0., 1., 2., 1., 2.,
            ...      0., 0., 2., 2., 0.]
            >>> u = numpy.array(u)
            >>> grad = rmg.calculate_gradients_at_active_links(u)
            >>> grad
            array([ 1.,  1., -1., -1., -1., -1., -1.,  0.,  1.,  1.,  1., -1.,  1.,
                    1.,  1., -1.,  1.])
            >>> flux = -grad    # downhill flux proportional to gradient
            >>> divflux = rmg.calculate_flux_divergence_at_active_cells(flux)
            >>> divflux
            array([ 2.,  4., -2.,  0.,  1., -4.])
            
        If calculate_gradients_at_active_links is called inside a loop, you can
        improve speed slightly by creating an array outside the loop. For 
        example, do this once, before the loop:
            
            >>> divflux = rmg.create_active_cell_dvector() # outside loop
            
        Then do this inside the loop:
            
            >>> divflux = rmg.calculate_flux_divergence_at_active_cells(flux, divflux)
            
        In this case, the function will not have to create the divflux array.
            
        """
        
        if self.DEBUG_TRACK_METHODS:
            print 'RasterModelGrid.calculate_flux_divergence_at_active_cells'
            
        assert (len(active_link_flux)==self.num_active_links), \
               "incorrect length of active_link_flux array"
            
        # If needed, create net_unit_flux array
        if net_unit_flux is False:
            net_unit_flux = numpy.zeros(self.num_active_cells)
        else:
            net_unit_flux[:] = 0.
            
        assert (len(net_unit_flux))==self.num_active_cells
        
        node_net_unit_flux = self.calculate_flux_divergence_at_nodes(active_link_flux)
                
        net_unit_flux = node_net_unit_flux[self.activecell_node]
        
        return net_unit_flux

    def calculate_max_gradients_at_nodes(self, link_gradients, max_gradient=False):
        """
        Created DEJH Sept 2013. Needs proper documentation. Based on approach of calc_flux_divergence..., below.
        NOT YET FUNCTIONAL.
        """
        
        if self.DEBUG_TRACK_METHODS:
            print 'RasterModelGrid.calculate_max_gradients_at_nodes'
            
        assert (len(link_gradients)==self.num_links), \
               "incorrect length of active_link_flux array"

        # If needed, create max_gradient array
        if max_gradient is False:
            max_gradient = numpy.zeros(self.num_nodes)
        else:
            max_gradient[:] = 0.

        assert(len(max_gradient) == self.num_nodes)

        gradients = numpy.zeros(len(link_gradients)+1)
        gradients[:-1] = link_gradients
        
        #Extract the values for each node_inlink/outlink_matrix, remembering to make the inlinks negative.
        #Search for the maxima.
        #Introduce a way of working on the diagonals!!!
        
        gradients[self.node_active_outlink_matrix[0][:]]


    def calculate_flux_divergence_at_nodes(self, active_link_flux, 
                                           net_unit_flux=False):
        """
        Same as calculate_flux_divergence_at_active_cells, but works with and
        returns a list of net unit fluxes that corresponds to all nodes, rather
        than just active cells.
        
        Note that we DO compute net unit fluxes at boundary nodes (even though
        these don't have active cells associated with them, and often don't have 
        cells of any kind, because they are on the perimeter). It's up to the 
        user to decide what to do with these boundary values.
        
        Example:
            
            >>> rmg = RasterModelGrid(4, 5, 1.0)
            >>> u = [0., 1., 2., 3., 0.,
            ...      1., 2., 3., 2., 3.,
            ...      0., 1., 2., 1., 2.,
            ...      0., 0., 2., 2., 0.]
            >>> u = numpy.array(u)
            >>> grad = rmg.calculate_gradients_at_active_links(u)
            >>> grad
            array([ 1.,  1., -1., -1., -1., -1., -1.,  0.,  1.,  1.,  1., -1.,  1.,
                    1.,  1., -1.,  1.])
            >>> flux = -grad    # downhill flux proportional to gradient
            >>> df = rmg.calculate_flux_divergence_at_nodes(flux)
            >>> df
            array([ 0., -1., -1.,  1.,  0., -1.,  2.,  4., -2.,  1., -1.,  0.,  1.,
                   -4.,  1.,  0., -1.,  0.,  1.,  0.])
            
        If calculate_gradients_at_nodes is called inside a loop, you can
        improve speed by creating an array outside the loop. For example, do
        this once, before the loop:
            
            >>> df = rmg.create_node_dvector() # outside loop
            >>> rmg.num_nodes
            20
            
        Then do this inside the loop:
            
            >>> df = rmg.calculate_flux_divergence_at_nodes(flux, df)
            
        In this case, the function will not have to create the df array.
        """
        
        if self.DEBUG_TRACK_METHODS:
            print 'RasterModelGrid.calculate_flux_divergence_at_nodes'
            
        assert (len(active_link_flux)==self.num_active_links), \
               "incorrect length of active_link_flux array"
            
        # If needed, create net_unit_flux array
        if net_unit_flux is False:
            net_unit_flux = numpy.zeros(self.num_nodes)
        else:
            net_unit_flux[:] = 0.
            
        assert(len(net_unit_flux) == self.num_nodes)
        
        flux = numpy.zeros(len(active_link_flux)+1)
        flux[:len(active_link_flux)] = active_link_flux * self._dx
        
        net_unit_flux = ((flux[self.node_active_outlink_matrix[0][:]] + \
                          flux[self.node_active_outlink_matrix[1][:]]) - \
                         (flux[self.node_active_inlink_matrix[0][:]] + \
                          flux[self.node_active_inlink_matrix[1][:]])) / self.cellarea

        return net_unit_flux
        
    def calculate_flux_divergence( self, q, id ):
        """
        ..todo: UPDATE THIS TO USE NEW DATA STRUCTURES!
        
        This is like calculate_flux_divergences (plural!), but only does
        it for cell "id".
        """
    
        if self.DEBUG_TRACK_METHODS:
            print 'RasterModelGrid.calculate_flux_divergence here with cell',id
            print 'q:',q[self.faces[id,0:4]]
        fd = ( -( q[self.faces[id,2]]   # left face (positive=in)
            + q[self.faces[id,3]] )           # bottom face (positive=in)
            + q[self.faces[id,0]]             # right face (positive=out)
            + q[self.faces[id,1]]             # top face (positive=out)
              ) / self._dx
        return fd
        
    def update_noflux_boundaries( self, u, bc = None ):
        """
        Sets the value of u at all noflux boundary cells equal to the
        value of their interior neighbors, as recorded in the
        "boundary_nbrs" array.

        .. deprecated:: 0.1
            Use `update_boundaries` instead
        """
    
        if bc==None:
            bc = self.default_bc

        inds = (bc.boundary_code[id] == bc.TRACKS_CELL_BOUNDARY)
        u[self.boundary_cells[inds]] = u[bc.tracks_cell[inds]]

        return u

    def update_boundaries( self, u, bc = None ):
        """
        Updates FIXED_GRADIENT and TRACKS_CELL boundaries in 
        BoundaryCondition "bc" (by default ModelGrid's default_bc), so
        that TRACKS_CELL values in u are assigned the value at the 
        corresponding cell-to-track, and FIXED_GRADIENT cells get a value
        equal to the cell-to-track's value plus gradient times distance.

        .. note:: Does NOT change fixed-value boundary cells.

        .. todo::
            use the cell-local distance rather than dx, for use in the base
            class!
            probably now obsolete (GT Aug 2013)
            
        NG Wondering if this should be boundary nodes, not cells.    
        """
    
        if bc == None:
            bc = self.default_bc

        inds = (bc.boundary_code == bc.TRACKS_CELL_BOUNDARY)
        u[self.boundary_cells[inds]] = u[bc.tracks_cell[inds]]

        inds = (bc.boundary_code == bc.FIXED_GRADIENT_BOUNDARY)
        u[self.boundary_cells[inds]] = (u[bc.tracks_cell[id]] +
                                        bc.gradient[id] * self._dx)

        return u

    def node_vector_to_raster(self, u, flip_vertically=False):
        """
        Converts node vector *u* to a 2D array and returns it, so that it
        can be plotted, output, etc.
        
        If the *flip_vertically* keyword is True, this function returns an 
        array that has the rows in reverse order. This is useful for use in
        plot commands (such as the image display functions) that puts the 
        first row at the top of the image. In the landlab coordinate system,
        the first row is thought to be at the bottom. Thus, a flipped matrix
        will plot in the landlab style with the first row at the bottom.

        The returned array is a view of *u*, not a copy.
        
        Example:
            
        >>> rmg = RasterModelGrid(4, 5, 1.0)
        >>> u = rmg.create_node_dvector()
        >>> u = u + range(0, len(u))
        >>> u
        array([  0.,   1.,   2.,   3.,   4.,   5.,   6.,   7.,   8.,   9.,  10.,
                11.,  12.,  13.,  14.,  15.,  16.,  17.,  18.,  19.])
        >>> ur = rmg.node_vector_to_raster(u)
        >>> ur
        array([[  0.,   1.,   2.,   3.,   4.],
               [  5.,   6.,   7.,   8.,   9.],
               [ 10.,  11.,  12.,  13.,  14.],
               [ 15.,  16.,  17.,  18.,  19.]])
        >>> ur = rmg.node_vector_to_raster(u, flip_vertically=True)        
        >>> ur
        array([[ 15.,  16.,  17.,  18.,  19.],
               [ 10.,  11.,  12.,  13.,  14.],
               [  5.,   6.,   7.,   8.,   9.],
               [  0.,   1.,   2.,   3.,   4.]])
        """
        return sgrid.reshape_array(self.shape, u,
                                   flip_vertically=flip_vertically)

    def cell_vector_to_raster(self, u, flip_vertically=False):
        """
        Converts cell (i.e., interior node) vector u to a 2D array and returns it, 
        so that it can be plotted, output, etc.
        
        If the optional argument flip_vertically=True, the function returns an 
        array that has the rows in reverse order, for use in plot commands (such
        as the image display functions) that put the (0,0) axis at the top left 
        instead of the bottom left.
        
        Example:
            
            >>> rmg = RasterModelGrid(4, 5, 1.0)
            >>> u = rmg.create_cell_dvector()
            >>> u = u + range(0, len(u))
            >>> u
            array([ 0.,  1.,  2.,  3.,  4.,  5.])
            >>> ur = rmg.cell_vector_to_raster(u)
            >>> ur
            array([[ 0.,  1.,  2.],
                   [ 3.,  4.,  5.]])
            >>> ur = rmg.cell_vector_to_raster(u, flip_vertically=True)        
            >>> ur
            array([[ 3.,  4.,  5.],
                   [ 0.,  1.,  2.]])
        """
        return sgrid.reshape_array((self.shape[0] - 2, self.shape[1] - 2),
                                   u, flip_vertically=flip_vertically)

    def get_neighbor_list(self, *args):
        """get_neighbor_list([ids])

        Return lists of neighbor nodes for nodes with given *ids*. If *ids*
        is not given, return the neighbors for all of the nodes in the grid.
        For each node, the list gives neighbor ids as [right, top, left,
        bottom]. Set all neighbors of boundary nodes to -1.
        
        >>> mg = RasterModelGrid(4, 5)
        >>> mg.get_neighbor_list([-1, 6])
        array([[-1, -1, -1, -1],
               [ 7, 11,  5,  1]])
        >>> mg.get_neighbor_list(7)
        array([ 8, 12,  6,  2])

        ..todo: could use inlink_matrix, outlink_matrix
        """
        if self.neighbor_list_created == False:
            self.create_neighbor_list()

        if len(args) == 0:
            return self.neighbor_nodes
        elif len(args) == 1:
            return self.neighbor_nodes[args[0], :]
        else:
            raise ValueError('only zero or one arguments accepted')

    def create_neighbor_list( self ):
        """
        Creates a list of IDs of neighbor nodes for each node, as a
        2D array. Only interior nodes are assigned neighbors; boundary
        nodes get -1 for each neighbor. The order of the neighbors is [right,
        top, left, bottom].

        .. note:: This is equivalent to the neighbors of all cells,
            and setting the neighbors of boundary-node cells to -1. In such a
            case, each node has one cell and each node-cell pair have the
            same ID. However, this is the old-style grid structure as
            boundary nodes no longer have associated cells.

        .. todo: could use inlink_matrix, outlink_matrix

        .. todo: Change to use BAD_INDEX_VALUE instead of -1.
        """
        # DH created this.  NG only changed labels.
        assert(self.neighbor_list_created == False)

        self.neighbor_list_created = True 
        self.neighbor_nodes = sgrid.neighbor_node_array(
            self.shape, out_of_bounds=-1, boundary_node_mask=-1)
                
    def has_boundary_neighbor(self, ids):
        """
        Checks to see if one of the eight neighbor nodes of node(s) with
        *id* has a boundary node.  Returns True if a node has a boundary node,
        False if all neighbors are interior.

        >>> mg = RasterModelGrid(5, 5)
        >>> mg.has_boundary_neighbor(6)
        True
        >>> mg.has_boundary_neighbor(12)
        False
        >>> mg.has_boundary_neighbor([12, -1])
        array([False,  True], dtype=bool)

        >>> mg.has_boundary_neighbor(25)
        Traceback (most recent call last):
            ...
        IndexError: index 25 is out of bounds for axis 0 with size 25
        """
        ans = has_boundary_neighbor(self, ids)
        if ans.ndim == 0:
            return bool(ans)
        else:
            return ans

    def get_diagonal_list(self, *args):
        """get_diagonal_list([ids])

        Return lists of diagonals nodes for nodes with given *ids*. If *ids*
        is not given, return the diagonals for all of the nodes in the grid.
        For each node, the list gives diagonal ids as [topright, topleft,
        bottomleft, bottomright]. Set all diagonals for boundary nodes to -1.
        
        >>> mg = RasterModelGrid(4, 5)
        >>> mg.get_diagonal_list([-1, 6])
        array([[-1, -1, -1, -1],
               [12, 10,  0,  2]])
        >>> mg.get_diagonal_list(7)
        array([13, 11,  1,  3])

        ..todo: could use inlink_matrix, outlink_matrix
        """
        #Added DEJH 051513
    
        if self.diagonal_list_created==False:
            self.create_diagonal_list()
        
        if len(args) == 0:
            return self.diagonal_cells
        elif len(args) == 1:
            return self.diagonal_cells[args[0], :]
        else:
            raise ValueError('only zero or one arguments accepted')

    def create_diagonal_list(self):
        """
        Creates a list of IDs of the diagonal nodes to each node, as a 2D
        array.  Only interior nodes are assigned diagonal neighbors; boundary
        nodes get -1 for each neighbor. The order of the diagonal nodes is
        [topright, topleft, bottomleft, bottomright].
        
        .. note:: This is equivalent to the diagonals of all cells,
            and setting the neighbors of boundary-node cells to -1. In such a
            case, each node has one cell and each node-cell pair have the
            same ID. However, this is the old-style grid structure as
            boundary nodes no longer have associated cells.

        .. todo: Change to use BAD_INDEX_VALUE instead of -1.
        """
        assert(self.diagonal_list_created == False)

        self.diagonal_list_created = True
        self.diagonal_cells = sgrid.diagonal_node_array(
            self.shape, out_of_bounds=-1, boundary_node_mask=-1)

    def is_interior( self, id ):
        """
        Returns True if the node is an interior node, False otherwise. 
        Interior status is indicated by a value of 0 in node_status.
        """
        # NG changed this.
        #ng thinks there may be a problem here.
        #return self.boundary_ids[id] < 0
        #return self.node_status[id] < 1
        return self.node_status[id] == INTERIOR_NODE
        
    def get_boundary_code( self, id ):
        """
        Returns the boundary status of a node.
        """
        # ng june 2013
        return self.node_status[id] 
        
    def get_face_connecting_cell_pair(self, cell_a, cell_b):
        """
        Returns an array of face indices that *cell_a* and *cell_b* share.
        If the cells do not share any faces, returns an empty array.
        """
        cell_faces = self.cell_faces([cell_a, cell_b])
        return numpy.intersect1d(cell_faces[0], cell_faces[1],
                                 assume_unique=True)

    def top_edge_node_ids(self):
        """
        Returns a 1D numpy integer array containing the node ID numbers of the 
        nodes along the top (y=ymax) grid edge.
        
        Example:
            
            >>> rmg = RasterModelGrid(4, 5, 1.0)
            >>> rmg.top_edge_node_ids()
            array([15, 16, 17, 18, 19])
        """
        return sgrid.top_edge_node_ids(self.shape)
        
    def bottom_edge_node_ids(self):
        """
        Returns a 1D numpy integer array containing the node ID numbers of the 
        nodes along the bottom (y=0) grid edge.
        
        Example:
            
            >>> rmg = RasterModelGrid(4, 5, 1.0)
            >>> rmg.bottom_edge_node_ids()
            array([0, 1, 2, 3, 4])
        """
        return sgrid.bottom_edge_node_ids(self.shape)
        
    def left_edge_node_ids(self):
        """
        Returns a 1D numpy integer array containing the node ID numbers of the 
        nodes along the left (x=0) grid edge.
        
        Example:
            
            >>> rmg = RasterModelGrid(4, 5, 1.0)
            >>> rmg.left_edge_node_ids()
            array([ 0,  5, 10, 15])
        """
        return sgrid.left_edge_node_ids(self.shape)
        
    def right_edge_node_ids(self):
        """
        Returns a 1D numpy integer array containing the node ID numbers of the 
        nodes along the right (x=xmax) grid edge.
        
        Example:
            
            >>> rmg = RasterModelGrid(4, 5, 1.0)
            >>> rmg.right_edge_node_ids()
            array([ 4,  9, 14, 19])
        """
        return sgrid.right_edge_node_ids(self.shape)
        
    def grid_coords_to_node_id(self, row, col, **kwds):
        """
        Returns the ID of the node at the specified *row* and *col* of the
        raster grid. Since this is a wrapper for the numpy ravel_multi_index
        function, the keyword arguments are the same as that function. In
        addition, *row* and *col* can both be either scalars or arrays (of the
        same length) to get multiple ids.

        As with ravel_multi_index use the *mode* keyword to change the
        behavior of the method when passed an out-of-range *row* or *col*.
        The default is to raise ValueError (not IndexError, as you might
        expect).
        
        .. note::
            The syntax assumes that first row and column are 0,
            so max entry for a mg with 4 rows and 5 cols is row=3, col=4
        
        Example:
            
            >>> mg = RasterModelGrid(4, 5)
            >>> mg.grid_coords_to_node_id(2, 3)
            13

            >>> mg.grid_coords_to_node_id([2, 0], [3, 4])
            array([13,  4])
        """
        return numpy.ravel_multi_index((row, col), self.shape, **kwds)
        
    def unit_test( self ):
        """
        This is just scratch space for testing while developing. More proper
        tests are in the doctests for each function, and in  
        test_raster_model_grid.py.
        """
        
        print 'Performing some tests for RasterModelGrid ...'
        print
        
        num_rows_for_unit_test = 4
        num_cols_for_unit_test = 5
        
        print 'Initializing ...'
        self.initialize( num_rows_for_unit_test, 
                         num_cols_for_unit_test,
                         1.0 )
        print 'done.'
        print
        
        #print 'Testing node lists:'
        #print 'ID   X    Y    Z    Status  Active_cell  #in in1 in2 #out out1 out2'
        #for node in range( 0, self.num_nodes ):
        #    print(str(node)+'    '+str(self._node_x[node])+'  '
        #          +str(self._node_y[node])+'  '
        #          +str(self._node_z[node])+'  '
        #          +str(self.node_status[node])+'  '
        #          +str(self.node_activecell[node])+'  '
        #          +str(self.node_numinlink[node])+'  '
        #          +str(self.node_inlink_matrix[0][node])+'  '
        #          +str(self.node_inlink_matrix[1][node])+'  '
        #          +str(self.node_numoutlink[node])+'  '
        #          +str(self.node_outlink_matrix[0][node])+'  '
        #          +str(self.node_outlink_matrix[1][node])+'  '
        #          +str(self.node_numactiveinlink[node])+'  '
        #          +str(self.node_active_inlink_matrix[0][node])+'  '
        #          +str(self.node_active_inlink_matrix[1][node])+'  '
        #          +str(self.node_numactiveoutlink[node])+'  '
        #          +str(self.node_active_outlink_matrix[0][node])+'  '
        #          +str(self.node_active_outlink_matrix[1][node]))
        #print
        #
        
        print 'Testing fluxes and in/out links:'
        flux2 = numpy.zeros(len(flux)+1)
        flux2[:len(flux)] = flux
        print flux2
        for n in range(0, self.num_nodes):
            mysum = -(flux2[self.node_active_inlink_matrix[0][n]] + \
                      flux2[self.node_active_inlink_matrix[1][n]]) + \
                     (flux2[self.node_active_outlink_matrix[0][n]] + \
                      flux2[self.node_active_outlink_matrix[1][n]])
            print(str(n)+' '+str(flux2[self.node_active_inlink_matrix[0][n]])
                        +' '+str(flux2[self.node_active_inlink_matrix[1][n]])
                        +' '+str(flux2[self.node_active_outlink_matrix[0][n]])
                        +' '+str(flux2[self.node_active_outlink_matrix[1][n]])
                        +' '+str(mysum))
        divg2 = self.calculate_flux_divergence_at_nodes(flux)
        print divg2


def _is_closed_boundary(boundary_string):
    
    return boundary_string.lower() == 'closed'


def from_dict(param_dict):
    """
    Create a RasterModelGrid from the dictionary-like object, *param_dict*.
    Required keys of the dictionary are NUM_ROWS, NUM_COLS. Raises a KeyError
    if either of these are missing.  If GRID_SPACING is given, use it as the
    HexModelGrid *dx* parameter, otherwise default to unit spacing.
    """
    # Read and create basic raster grid
    try:
        nrows = int(param_dict['NUM_ROWS'])
        ncols = int(param_dict['NUM_COLS'])
        dx = float(param_dict.get('GRID_SPACING', 1.))
    except KeyError:
        raise
    except ValueError:
        raise
    else:
        mg = RasterModelGrid(nrows, ncols, dx)
        
    # Set boundaries
    left_boundary_type = param_dict.get('LEFT_BOUNDARY', 'open')
    right_boundary_type = param_dict.get('RIGHT_BOUNDARY', 'open')
    top_boundary_type = param_dict.get('TOP_BOUNDARY', 'open')
    bottom_boundary_type = param_dict.get('BOTTOM_BOUNDARY', 'open')
    mg.set_inactive_boundaries(_is_closed_boundary(bottom_boundary_type), 
                               _is_closed_boundary(right_boundary_type),
                               _is_closed_boundary(top_boundary_type),
                               _is_closed_boundary(left_boundary_type))

    # Return the created and initialized grid
    return mg



#! /usr/bin/env python
"""
The Landlab

:Package name: TheLandlab
:Version: 0.1.0
:Release date: 2013-03-24
:Authors:
  Greg Tucker,
  Nicole Gasparini,
  Erkan Istanbulluoglu,
  Daniel Hobley,
  Sai Nudurupati,
  Jordan Adams,
  Eric Hutton

:URL: http://csdms.colorado.edu/trac/landlab

:License: MIT
"""
from landlab.model_parameter_dictionary import ModelParameterDictionary
from landlab.framework.collections import Palette, Arena, NoProvidersError
from landlab.framework.decorators import Implements, ImplementsOrRaise
from landlab.framework.framework import Framework
from landlab.grid import *
from landlab.model_component import Component



#! /usr/bin/env python

class Component(object):
    _input_var_names = set()
    _output_var_names = set()
    _var_units = dict()

    def __init__(self, grid, map_vars={}):
        self._grid = grid

        for (dest, src) in map_vars.items():
            grid[dest] = grid[src]

    @property
    def input_var_names(self):
        return self._input_var_names

    @property
    def output_var_names(self):
        return self._output_var_names

    @property
    def name(self):
        return self._name

    @property
    def units(self):
        return self._var_units

    @property
    def shape(self):
        return self.grid._shape

    @property
    def grid(self):
        return self._grid

    @property
    def coords(self):
        return (self.grid.get_cell_x_coords(), self.grid.get_cell_y_coords())

    def imshow(self, name, **kwds):
        self._grid.imshow(name, **kwds)



#! /usr/env/python

from random import random
import math
import numpy
from collections import deque
import sys
import time
import scipy.optimize as opt
from landlab import RasterModelGrid #this is the tMesh equivalent module

#these ones only so we can run this module ad-hoc:
from pylab import plot, draw, show, contour, imshow, colorbar
from copy import copy

sys.setrecursionlimit(1500)

class data(object):
    '''
    This is where all the whole-grid data lives, as arrays over the various elements of the grid.
    '''
    #Data goes here!!!
    def __init__(self, grid):
        self.elev = grid.create_node_dvector() #some data
        self.flag_already_in_the_list = grid.create_node_dvector()
        self.craters_over_max_radius_not_plotted = grid.create_node_dvector()
        self.impact_sequence = []

class impactor(object):
    '''
    This class holds all parameters decribing properties of a single impact structure, and contains methods for recalculating fresh and internally consistent data describing such a impact structure.
    Built DEJH Spring 2013.
    '''
    def __init__(self, min_radius=0.005):
        self._xcoord = -999.
        self._ycoord = -999.
        self._str_regime_cutoff = 0.3 #??? #Holsapple gives 0.1km as the cutoff for R/d scaling, Moore in Housen gives this as the likely Lunar cutoff for true str regime (p. 2495 Housen)
        self._simple_radius_depth_ratio_Pike = 2.55 #Pike thru Holsapple, tho Garvin (2011) gives 2.0 as "typically cited"
        #the ratio for complex craters is called as a method, and not set as such
        self._simple_cr_radius_cutoff = 7.5 #km following Melosh p. 224 - LUNAR
        self._complex_cr_radius_cutoff = 10. #km, folling Melosh, i.e., descending
    #anything in between these values needs a chance of being either - Pike says transition range is R = 6-12.5 km
        self._max_complex_cr_radius = 70. #km, bigger is a "peak ring crater". This is not currently used.
        #self._power_for_complex_crater = 2. #i.e., d = D(r/R)^power. =2 is a parabola. This is no longer used; replaced by get_crater_shape_exp() for all craters simple and complex.
        #we follow Garvin et al 2011 and assume power=1.45 for simple crater
        #NB - if we assume a power law and dictate S=32deg at the rim, we recover n=0.51R/D - gives n~1.3 for Linne (synthetic; if real depth it's ~1) and maxes out ~2 for biggest craters. (Plus infill by lava?) ->This is a nice way to do it. This parameter is presently unused.
        self._minimum_crater = min_radius #km. This is the smallest modelled crater radius. 10m diameter is the strict cutoff for known cr distns
        self.ivanov_a = [-3.0876, -3.557528, 0.781027, 1.021521, -0.156012, -0.444058, 0.019977, 0.08685, -0.005874, -0.006809, 0.000825, 0.0000554] #The coefficients for Ivanov's crater distn function
        self._impactor_angle_to_surface = -999.
        self._angle_to_horizontal = -999.
        self._minimum_ejecta_thickness = 0.000001
        #NB - If this min thickness is changed, the optimization point in excavate_a_crater_optimized() will also need to be changed
        self._beta_factor = 0.5 #this is the arbitrary term that controls how "stretched out" the ejecta field is. <+0.5 prevents "outside the ejecta field" regions forming
        
        self.total_counted_craters = self.ivanov_prod_equ_as_Nequals(self._minimum_crater*2.)
        #Define the Ivanov fn and its derivatives, for use in Newton Raphson optimization. Note this is normalized now.
        self.ivanov_prod_fn = lambda x, N_as_fraction: self.ivanov_prod_equ(x,(N_as_fraction/self.total_counted_craters))
        self.ivanov_prod_fn_1stderiv = lambda x, N_as_fraction: self.ivanov_prod_equ_1stderiv(x, N_as_fraction)
        #self.ivanov_prod_fn_2ndderiv = lambda x: self.ivanov_prod_equ_2ndderiv(x)

    
    def get_complex_radius_depth_ratio(self):
        '''
        This method returns the ratio of radius to max depth for a complex crater, which is itself a function of that crater's radius. This equation is from Pike '77, as quoted in Holsapple '93, p. 358.
        '''
        return self._radius**0.7 #Pike 77, following Holsapple 93, p. 358 - Not clear how I need to call this again to get it to update before use
    
    
    def get_crater_shape_exp(self):
        '''
        This method assumes the max depth and radius of a crater are known. It provides n for a power law of form d = D*(r/R)**n, where D and R are the known values, by assuming the outer edges of the crater sit at angle of repose. This gives very sensible answers; n~2 for big, complex craters (Garvin et al, 2000, p.333: "There is a strong tendency for craters to become more paraboloidal with increasing diameter, independent of location.") and n~1.3 for ~2km simple craters (Garvin following Croft has ~1.18).
        '''
        return 0.51 * self._radius / self._depth
    
    
    def get_complex_peak_radius(self):
        '''
        This method returns the radial distance to the edge of a complex crater central mound, based on the crater radius. This follows an equation presented in Melosh's Planetary Surface Processes.
        '''
        return 0.22 * self._radius #Melosh
    
    
    def get_complex_peak_str_uplift(self):
        '''
        This method returns the radial maximum structural uplift, interpreted as a height, of a complex crater central mound, based on the crater radius. This follows an equation presented in Melosh's Planetary Surface Processes, adjusted for R, not D.
        '''
        return 0.13 * self._radius ** 1.1 #melosh, note correction for R not D
    
    
    #Holsapple & Housen et al 1983 note that in the strength regime, ejecta distribution will NOT scale independently of crater size (it does in the gravity regime) - ejecta will be proportionally FURTHER FROM the rim as craters get smaller (e.g., Housen et al 83, fig 9)
    #...this, of course, makes small craters disproportionately effective diffusers - must find a way of quantifying this!
    #Moore et al 74 cited in Housen et al 83 gives Rce/R as const ~2.35 +0.56 -0.45 std err for Lunar craters 0.3-100km, where Rce is the continuous ejecta limit. Below this, seems to level off abruptly - i.e., Rce ~ R**(<1) (and const will change also to maintain connection at 0.3km)
    #A key point - geomorphic gardening is not totally comparable to regolith gardening. We care about topology - the final crater. Regolith fracture and turnover is more likely to depend on the crater transient depth.
    
    
    def ivanov_prod_equ(self, D, N):
        '''
        This function expresses the crater production function after Ivanov, 1999, in the form 0 = f(D, N). N is the total number density of craters greater than x diameter. Remember, N is in km^-2 Ga^-1. If N = N_counted / N_max, then this is the CPF which will yield the PDF when solved for D given N.
        '''
        sum_terms = numpy.empty(12, dtype=float)
        sum_terms[0] = self.ivanov_a[0] - math.log10(N)
        for i in range(1,12):
            sum_terms[i] = self.ivanov_a[i] * math.log10(D) ** i
        return numpy.sum(sum_terms)
    
    
    #A second version, in the conventional form N = f(D):
    def ivanov_prod_equ_as_Nequals(self, D):
        '''
        This function expresses Ivanov's CPF in the normal form, N = f(D). It returns N (not log10(N)!).
        '''
        sum_terms = numpy.empty(12, dtype=float)
        for i in range (0,12):
            sum_terms[i] = self.ivanov_a[i] * math.log10(D) ** i
        return pow( 10., numpy.sum(sum_terms) )

    
    def ivanov_prod_equ_1stderiv(self, D, N):
        '''
        This is a helper function for the Newton-Raphson optimized solver in solve_ivanov_for_crater_diam. It returns the first derivative for the production function. Note that the N term is a dummy variable used for consistency in the opt.newton(), and does nothing here.
        '''
        sum_terms = numpy.empty(11, dtype=float)
        ln_of_10 = math.log(10)
        for i in range(1,12):
            sum_terms[i-1] = self.ivanov_a[i] * i * math.log(D) ** (i-1.) / ln_of_10**i
        return (numpy.sum(sum_terms) / D)


    def ivanov_prod_equ_2ndderiv(self, D):
        '''
        This is a helper function for the Newton-Raphson optimized solver in solve_ivanov_for_crater_diam. It returns the second derivative for the production function.
        '''
        ln_of_10 = math.log(10)
        sum_terms = numpy.empty(11, dtype=float)
        sum_terms[0] = self.ivanov_a[1] / ln_of_10
        for i in range(2,12):
            sum_terms[i-1] = i * self.ivanov_a[i] * (math.log(D) - i + 1.) * math.log(D)**(i-2.) / ln_of_10**i
        return (-numpy.sum(sum_terms) / (D*D))


    def solve_ivanov_for_crater_diam(self, N_by_fract):
        '''
        Uses a Newton Raphson method to return a random diameter drawn from an Ivanov crater distribution, when provided with a uniformly distributed random number 0->1.
        '''
        args_in = N_by_fract,
        #Estimate zero position by a super ad-hoc Shoemaker distn. Note K is not really right here - we've defined it at D=1; I think it should be equal to Ivanov at D~0.03.
        x_0 = (10.**self.ivanov_a[0] / (self.total_counted_craters * N_by_fract)) ** 0.345
        return opt.newton(func=self.ivanov_prod_fn, x0=x_0, fprime=self.ivanov_prod_fn_1stderiv, args=args_in, tol=0.001)
    
    
    def set_cr_radius_from_shoemaker(self, data):
        '''
        This method is a less accurate (but faster) alternative to the Newton-Raphson-on-Ivanov-distn also available in this object. It takes a random number between 0 and 1, and returns a crater radius based on a py distn N = kD^-2.9, following Shoemaker et al., 1970.
        '''
        self._radius = self._minimum_crater*(random())**-0.345
        if self._radius > self._max_complex_cr_radius:
            data.craters_over_max_radius_not_plotted.append(self._radius)
            print 'Drew a crater above the maximum permitted size. Drawing a new crater...'
            self.set_size(data)


    def set_coords(self, grid, data):
        '''
        This method selects a random location inside the grid onto which to map an impact. It also sets variables for the closest grid node to the impact, and the elevation at that node.
        '''
        #NB - we should be allowing craters OUTSIDE the grid - as long as part of them impinges.
        #This would be relatively easy to implement - allow allocation out to the max crater we expect, then allow runs using these coords on our smaller grid. Can save comp time by checking if there will be impingement before doing the search.
        self._xcoord = random() * (grid.get_grid_xdimension() - grid.dx)
        self._ycoord = random() * (grid.get_grid_ydimension() - grid.dx)
        #Snap impact to grid:
        vertices_array = grid.get_nodes_around_point(self._xcoord, self._ycoord)
        distances_to_vertices = numpy.empty(vertices_array.size)
        counter = 0
        for x in vertices_array:
            #print 'The ID is:', x
            distances_to_vertices[counter] = numpy.sqrt(self._xcoord-grid.node_x[x]**2. + self._ycoord-grid.node_y[x]**2.)
            counter += 1
        self.closest_node_index = vertices_array[numpy.argmin(distances_to_vertices)]
        self.closest_node_elev = data.elev[self.closest_node_index]
        #print 'Closest node elev and index: ', self.closest_node_elev, self.closest_node_index
        # Check we haven't hit a boundary node. Reset if we have:
        #NB - this step can introduce recursion very easily into the program. We need a way out of this that doesn't result in endless recursion. 
        #if not grid.is_interior(self.closest_node_index):
            #self.set_coords(grid, data)
        #NB - snapping to the grid may be quite computationally demanding in a Voronoi. Could dramatically increase speed by picking a node for centre and working out from there (or just leaving it on the node...). Fine for square grid though.
            

    def set_impactor_angles(self):
        '''
        This method sets the angle of impact, assuming the only effect is rotation of the planet under the impactor bombardment (i.e., if the target looks like a circle to the oncoming impactor, there's more limb area there to hit). As long as target is rotating relative to the sun, other (directional) effects should cancel. Angle is given to horizontal. Also sets a random azimuth.
        '''
        self._angle_to_horizontal =  numpy.arccos(random()) #gives sin distn with most values drawn nearer to 0
        self._azimuth_of_travel = random() * 2. * numpy.pi #equal chance of any azimuth
        #Shoemaker 1983 gives a speculative equn for the suppression of D by increasing impact angle (his equ 3), but effect is minor, and it's probably not worth the trouble.
        #Shoemaker 1962 (in book) apparently states low angle impacts are very rare. Need to read this.

    
    def set_size(self, data):
        '''
        This method draws a crater radius at random from the PDF describing crater sizes dictated by the Ivanov distribution, with the probability of obtaining that crater size depending on the relative abundance of that size. Note this method works with the Ivanov distribution according to diameter, but returns a radius.
        '''
        self._radius = 0.5 * self.solve_ivanov_for_crater_diam(random())
        if self._radius > self._max_complex_cr_radius:
            data.craters_over_max_radius_not_plotted.append(self._radius)
            print 'Drew a crater above the maximum permitted size. Drawing a new crater...'
            self.set_size(data)

    
    def set_depth_from_size(self):
        '''
        This method sets a known crater with diameter as either simple or complex, then sets its maximum depth at the center (inferred for complex craters).
        '''
        #Needs to account for simple to complex.
        #Let's ignore the transition at the lowest size scales for now.
        #Need to consider the transition at the top end too!!
        if self._radius < self._simple_cr_radius_cutoff:
            self._crater_type = 0 #simple
        elif self._radius > self._complex_cr_radius_cutoff:
            self._crater_type = 1 #complex
        else:
            py_complex_at_radius = (self._radius - self._simple_cr_radius_cutoff) / (self._complex_cr_radius_cutoff - self._simple_cr_radius_cutoff)
            if random() > py_complex_at_radius:
                self._crater_type = 0 #simple
            else:
                self._crater_type = 1 #complex
                
        if not self._crater_type:
            self._depth = self._radius / self._simple_radius_depth_ratio_Pike
        else:
            self._depth = self._radius / self.get_complex_radius_depth_ratio()
            #Derive the scaling metrics for a complex crater, for subsequent use:
            self._complex_peak_radius = self.get_complex_peak_radius()
            self._complex_peak_str_uplift = self.get_complex_peak_str_uplift()

                
    def set_crater_volume(self):
        '''
        This method uses known crater depth and radius and sets the volume of the excavated cavity.
        '''
        self._cavity_volume = 0.51 * numpy.pi * self._depth * self._radius**3. / (0.51*self._radius + 2.*self._depth)


    def set_crater_mean_slope(self, grid, data):
        '''
        This method takes a crater of known radius, and which has already been "snapped" to the grid through snap_impact_to_grid(mygrid), and returns a spatially averaged value for the local slope of the preexisting topo beneath the cavity footprint. For computational efficiency reasons, the average slope is a mean of local slopes at 8 points 45 degrees apart around the crater rim, plus the center point. This function also  sets the mean surface dip direction.
            NOW DEPRECIATED, USE V2!
        '''
        #This might not be the best way to do this... There are issues with the magnitude of the slope being high if the individual points slope steeply even if they point in different directions.
        half_crater_radius = 0.707 * self._radius
        radial_points = []
        #elevation_array = []
        slope_array = []
        dip_direction_array = []
        #Work round clockwise from 12 o'clock
        slope_pts = [[self._xcoord,self._ycoord+self._radius],
                     [self._xcoord+half_crater_radius,self._ycoord+half_crater_radius],
                     [self._xcoord+self._radius,self._ycoord],
                     [self._xcoord+half_crater_radius,self._ycoord-half_crater_radius],
                     [self._xcoord,self._ycoord-self._radius],
                     [self._xcoord-half_crater_radius,self._ycoord-half_crater_radius],
                     [self._xcoord-self._radius,self._ycoord],
                     [self._xcoord-half_crater_radius,self._ycoord+half_crater_radius]]
        for x,y in slope_pts:
            if x < grid.get_grid_xdimension() and y < grid.get_grid_ydimension():
                radial_points.append(self.snap_coords_to_grid(grid, x, y))
        for a in radial_points:
            if grid.is_interior(a):
                slope, dir = grid.calculate_max_gradient_across_node(data.elev, a)
            #Make sure we stay on the grid!!:
                if not numpy.isnan(slope): #Probably not necessary now; retain for belt and braces
                    slope_array.append(slope)
                    dip_direction_array.append(dir)
                else:
                    print 'Slope is not defined at one of the nodes round the crater rim.'
            else:
                print 'One of the nodes round the crater rim is a boundary cell; no slope calculated here.'
        # Add in the central node:
        slope0, dir0 = grid.calculate_max_gradient_across_node(data.elev, self.closest_node_index)
        if not numpy.isnan(slope0):
            slope_array.append(slope0)
            dip_direction_array.append(dir0)
        self._surface_slope = numpy.mean(slope_array)
        dip_direction_x = []
        dip_direction_y = []
        for a in dip_direction_array:
            dip_direction_x.append(numpy.sin(a))
            dip_direction_y.append(numpy.cos(a))
        dip_direction_x_mean = numpy.mean(dip_direction_x)
        dip_direction_y_mean = numpy.mean(dip_direction_y)
        angle_to_yaxis = numpy.arctan(dip_direction_y_mean/dip_direction_x_mean)
        if dip_direction_x < 0.:
            self._surface_dip_direction = 1.5 * numpy.pi - angle_to_yaxis
        else:
            self._surface_dip_direction = 0.5 * numpy.pi - angle_to_yaxis
        print 'The mean slope is: ', self._surface_slope
    
    
    def set_crater_mean_slope_v2(self, grid, data):
        '''
        This method takes a crater of known radius, and which has already been "snapped" to the grid through snap_impact_to_grid(mygrid), and returns a spatially averaged value for the local slope of the preexisting topo beneath the cavity footprint. This version of the method works by taking four transects across the crater area every 45 degrees around its rim, calculating the slope along each, then setting the slope as the greatest, positive downwards and in the appropriate D8 direction. This function also sets the mean surface dip direction.
        '''
        #It would be trivial to add more divisions, i.e., D16, D32, D64, etc. Code is setup for this; just add to the slope_pts lists, and adjust the set length of the arrays with divisions
        divisions = 4 #len(radial_points1), i.e., half the no. of points == the number of lines across the wheel
        half_crater_radius = 0.707 * self._radius
        radial_points1 = numpy.empty(divisions, dtype=int)
        radial_points2 = numpy.empty(divisions, dtype=int)
        #elevation_array = []
        slope_array = []
        #dip_direction_array = []
        #Work round clockwise from 12 o'clock
        slope_pts1 = [[self._xcoord,self._ycoord+self._radius],
                     [self._xcoord+half_crater_radius,self._ycoord+half_crater_radius],
                     [self._xcoord+self._radius,self._ycoord],
                     [self._xcoord+half_crater_radius,self._ycoord-half_crater_radius]]
        slope_pts2 = [[self._xcoord,self._ycoord-self._radius],
                     [self._xcoord-half_crater_radius,self._ycoord-half_crater_radius],
                     [self._xcoord-self._radius,self._ycoord],
                     [self._xcoord-half_crater_radius,self._ycoord+half_crater_radius]]
        counter=0
        grid_xdimension = grid.get_grid_xdimension()
        grid_ydimension = grid.get_grid_ydimension()
        #this loop tests if the points round the rim are within the grid, then makes a list of nodes closest to that point. If not, it adds the central point to the list.
        for x,y in slope_pts1:
            if 0. < x < grid_xdimension and 0. < y < grid_ydimension:
                radial_points1[counter] = self.snap_coords_to_grid(grid, x, y)
            else:
                radial_points1[counter] = self.closest_node_index
            counter += 1
        counter=0
        for x,y in slope_pts2:
            if 0. < x < grid_xdimension and 0. < y < grid_ydimension:
                radial_points2[counter] = self.snap_coords_to_grid(grid, x, y)
            else:
                radial_points2[counter] = self.closest_node_index
            counter += 1
        for a in range(divisions):
            du = data.elev[radial_points1[a]] - data.elev[radial_points2[a]]
            #If du is negative, it means the surface slopes broadly EAST
            if radial_points1[a] is radial_points2[a]:
                dx = 0.
            elif radial_points1[a] is self.closest_node_index or radial_points2[a] is self.closest_node_index:
                dx = self._radius
            else:
                dx = 2.*self._radius
            if dx:
                slope_array.append(numpy.arctan(du/dx))
            else:
                slope_array.append(numpy.nan)
        slope_array = numpy.array(slope_array)
        #add test to see if crashing is due to problems in slope extraction:
        if numpy.isnan(slope_array).all():
            print 'All slopes beneath area did not evaluate! ABORT ABORT ABORT!!!'
        hi_mag_slope_index = numpy.argmax(numpy.fabs(slope_array))
        hi_mag_slope = slope_array[hi_mag_slope_index]
        #print hi_mag_slope
        if hi_mag_slope > 0.: #i.e., dips WEST (or dead S)
            self._surface_dip_direction = (hi_mag_slope_index/divisions + 1.)*numpy.pi
        elif not hi_mag_slope: #i.e., FLAT, dip dir is arbitrary
            self._surface_dip_direction = 0.
        else: #dips EAST
            self._surface_dip_direction = numpy.pi*hi_mag_slope_index/divisions    
        self._surface_slope = numpy.fabs(hi_mag_slope)
        print 'The slope under the crater cavity footprint is: ', self._surface_slope

        
    def snap_coords_to_grid(self, grid, xcoord, ycoord):
        '''
        This method takes existing coordinates, inside the grid, and returns the ID of the closest grid node.
        '''
            #There should probably be an exception here: return None?
        vertices_array = grid.get_nodes_around_point(self._xcoord, self._ycoord)
        distances_to_vertices = numpy.empty(len(vertices_array),dtype=float)
        counter = 0
        for x in vertices_array:
            distances_to_vertices[counter] = numpy.sqrt((xcoord-grid.node_x[x])**2. + (ycoord-grid.node_y[x])**2.)
            counter += 1
        return vertices_array[numpy.argmin(distances_to_vertices)]


    def set_elev_change_at_pts(self, grid, data):
        '''
        This is a method to take an existing impact properties and a known nearest node to the impact site, and alter the topography to model the impact. It assumes crater radius and depth are known, models cavity shape as a power law where n is a function of R/D, and models ejecta thickness as an exponential decay,sensitive to both ballistic range from tilting and momentum transfer in impact (after Furbish). We DO NOT yet model transition to peak ring craters, or enhanced diffusion by ejecta in the strength regime. Peak ring craters are rejected from the distribution. This version of this method is designed to remove the sheer walls around the edges of craters, and replace them with a true dipping rim.
        '''
        
        #Load in the data, for speed:
        _angle_to_horizontal=self._angle_to_horizontal
        _surface_slope=self._surface_slope
        _surface_dip_direction = self._surface_dip_direction
        _azimuth_of_travel = self._azimuth_of_travel
        pi = numpy.pi
        twopi = 2.*pi
        tan = numpy.tan
        cos = numpy.cos
        sin = numpy.sin
        sqrt = numpy.sqrt
        arctan = numpy.arctan
        _radius = self._radius
        elev = data.elev
        
        #Build a list of nodes to work on. Starts just with the node closest to the center.
        crater_node_list = deque([self.closest_node_index])
        elev_changes = deque()
        #Build an array of flags into the nodelist of the grid to note whether that node has been placed in the list this loop:
        flag_already_in_the_list = numpy.zeros(grid.get_count_of_all_nodes())
        
        #Derive the exponent for the crater shape, shared betw simple & complex:
        crater_bowl_exp = self.get_crater_shape_exp()
        print 'Crater shape exponent: ', crater_bowl_exp
        
        #print 'Radius is: ', _radius
        
        #Derive the effective angle for impact, relative to the surface, beta_eff. The direction is always controlled by the impactor. Draw new impact angles if impact geomtery is impossible.
        while 1:
            #This ugly function applies a correction to the azimuth of travel which will be the actual angle the ejecta is propelled along. This is important since if, e.g., _angle_to_horizontal ~pi/2, then the surface dip direction needs to be the dominant term.
            denominator = cos(_angle_to_horizontal) + sin(_surface_slope)*cos(_surface_dip_direction - _azimuth_of_travel)
            #SOURCE OF THE DIVBYZERO!!!
            #print 'Terms in denominator: ', cos(_angle_to_horizontal),  sin(_surface_slope), cos(_surface_dip_direction - _azimuth_of_travel)
            tan_angle_from_imp_az = sin(_surface_slope)*sin(_surface_dip_direction - _azimuth_of_travel) / denominator
            if denominator > 0.:
                _ejecta_azimuth = (_azimuth_of_travel + arctan(tan_angle_from_imp_az) + twopi)%(twopi)
            else:
                _ejecta_azimuth = (_azimuth_of_travel - arctan(tan_angle_from_imp_az) + pi)%(twopi)
            print 'Angle from impact azimuth: ', _ejecta_azimuth - _azimuth_of_travel
            beta_eff_options = [numpy.nan] * 2
            beta_eff_options[0] = _angle_to_horizontal*cos(_ejecta_azimuth - _azimuth_of_travel) - _surface_slope * cos(_ejecta_azimuth - _surface_dip_direction)
            beta_eff_options[1] = (pi - _angle_to_horizontal*cos(_ejecta_azimuth - _azimuth_of_travel)) + _surface_slope * cos(_azimuth_of_travel - _surface_dip_direction)
            beta_eff = min(beta_eff_options)
            #Need to think thru if this reversal is still required:
            #if beta_eff_options[0] > beta_eff_options[1]:
            #_azimuth_of_travel = (_azimuth_of_travel + pi)%(2.*pi)
            print 'Beta effective: ', beta_eff
            if beta_eff >= 0.:
                break
            else:
                print 'Refreshing the impactor angle'
                self.set_impactor_angles()
                _azimuth_of_travel = self._azimuth_of_travel
                _angle_to_horizontal = self._angle_to_horizontal
        _impactor_angle_to_surface_normal = 0.5*pi - beta_eff
        print 'Impact, ejecta azimuths: ', _azimuth_of_travel, _ejecta_azimuth
        #print 'Beta effective: ', beta_eff
        #rim_height = self._cavity_volume/(3.667*pi*_radius**2.)

        #Derive the new ejecta shape params:
        tan_repose = tan(32.*pi/180.)
        _b = 3.667*_radius*_radius*tan_repose
        _c = self._cavity_volume - 3.667*_radius*_radius*_radius*tan_repose - 0.5*_radius*_radius*tan_repose
        radius_calc = (-_b + sqrt(_b*_b - 2.*tan_repose*_c))/tan_repose
        thickness_at_rim = (_radius - radius_calc)*tan_repose
        #...where thickness(r) = thickness_at_rim*(r/R_true)**-2.75

        #This code crawls out iteratively over the grid under the crater footprint, away from the centerpoint. This could be made much faster in Python if it wasn't a loop. Maybe an alternative would be to define the radius corresponding to the minimum loop, then trim out a box 2*R by 2*R to operate on, and perform all actions on arrays for speed.
        while 1:
            try:
                active_node = crater_node_list.popleft() #i.e., array is not empty
            except:
                break
            else:
                active_node_x_to_center = grid.node_x[active_node] - self._xcoord
                active_node_y_to_center = grid.node_y[active_node] - self._ycoord
                active_node_r_to_center = sqrt(active_node_x_to_center*active_node_x_to_center + active_node_y_to_center*active_node_y_to_center)
                #Special case for if the impact location is right on a gridline:
                if not active_node_x_to_center:
                    if active_node_y_to_center < 0.:
                        active_node_theta = pi
                    else:
                        active_node_theta = 0.
                else:
                    active_node_angle_to_yaxis = arctan(active_node_y_to_center/active_node_x_to_center)
                    if active_node_x_to_center < 0.:
                        active_node_theta = 1.5 * pi - active_node_angle_to_yaxis
                    else:
                        active_node_theta = 0.5 * pi - active_node_angle_to_yaxis

                #In this version, our sweep out from the center will need to account for deposition depth elevating the crater rim, i.e., we need to deposit *before* we cut the cavity. We do this by defining three domains for the node to lie in: 1. r<r_calc, i.e., below the pre-impact surface. No risk of intersecting the surface here. 2. r_calc < r; Th>z_new. this is the domain in the inward sloping rim of the crater ejecta. 3. Th<z_new and beyond. out on the ejecta proper. Note - (1) is not hard & fast rule if the surface dips. Safer is just (Th-lowering)<z_new
                #So, calc the excavation depth for all nodes, just to be on the safe side for strongly tilted geometries:
                if active_node_r_to_center <= _radius:
                    new_z = self.closest_node_elev - self._depth * (1. - (active_node_r_to_center/_radius)**crater_bowl_exp) + thickness_at_rim
                else:
                    new_z = self.closest_node_elev + thickness_at_rim + (active_node_r_to_center-_radius)*tan_repose
                
                #print new_z

                if new_z<elev[active_node]: #Below the original surface
                    #print 'Under surface'
                    elev_changes.append(new_z-elev[active_node])
                    if self._crater_type: #Complex crater, adjust for the peak
                        if active_node_r_to_center <= self._complex_peak_radius: #on the peak
                            new_z  = new_z + self._complex_peak_str_uplift * (1. - active_node_r_to_center/self._complex_peak_radius)
                    #Set the ground elev:
                    elev[active_node] = new_z
                    #Add the neighboring nodes which haven't already been adjusted to the for-processing array:
                    neighbors_active_node = grid.get_neighbor_list(active_node)
                    for x in neighbors_active_node:
                        if not flag_already_in_the_list[x]:
                            if x!=-1: #Not an edge
                                crater_node_list.append(x)
                                flag_already_in_the_list[x] = 1

                else:
                    #Calc the ejecta thickness for a perp. impact. Note it's fine if r<R_true
                    local_flat_thickness = thickness_at_rim*(active_node_r_to_center/_radius)**-2.75
                    #print 'Local flat thickness: ', local_flat_thickness

                    # Modulating functions are f(impactor az, dip dir, impactor angle, dip), the ejecta mass term, and mu(impactor az, dip dir, impactor angle, dip), the distance travelled term. Momentum is the dominant term over geometric effects in mu, by ~an OoM - o just model as a momentum effect.
                    #Thickness = thickness_flat * f_theta / mu
                    #beta_eff (==pi/2 - angle from normal) is already known.
                    theta_eff = _ejecta_azimuth - active_node_theta #This is the angle of the center-to-active-node line to the azimuth along which the ejecta is concentrated
                    tan_beta = tan(_impactor_angle_to_surface_normal*self._beta_factor)
                    sin_theta_sqd = sin(theta_eff) ** 2.
                    cos_theta = cos(theta_eff)
                    #REMEMBER, as tan_beta gets >1, the function describing the ejecta is only valid over ever more restricted ranges of theta!! In other words,
                    if tan_beta and sin_theta_sqd*tan_beta*tan_beta > 1.:
                        thickness_at_active_node = 0.
                        #print 'Outside ejecta cone'
                    else:
                        mu_theta_by_mu0 = tan_beta * cos_theta + sqrt(1. - sin_theta_sqd * tan_beta*tan_beta)
                        f_theta = (tan_beta*tan_beta*(cos_theta*cos_theta-sin_theta_sqd) + 2.*tan_beta*cos_theta*sqrt(1.-tan_beta*tan_beta*sin_theta_sqd) + 1.) / twopi
                        #print f_theta/mu_theta_by_mu0
                        #So, distn_at_angle = distn_vertical_impact*f_theta/mu_theta_by_mu0. Draw the thickness at the active node:
                        #NB-the 2pi is to correct for mismatch in the dimensions of mu and f
                        thickness_at_active_node = f_theta/mu_theta_by_mu0 * twopi * local_flat_thickness
                        if thickness_at_active_node < 0.:
                            thickness_at_active_node = 0.
                    #Now, are we inside or outside the rim?
                    if new_z <= (elev[active_node] + thickness_at_active_node): #inside the rim
                        #print 'Inside the rim, on the ejecta'
                        elev_changes.append(new_z-elev[active_node])
                        elev[active_node] = new_z
                    else: #outside the rim
                        #print 'Outside the rim'
                        elev_changes.append(thickness_at_active_node)
                        elev[active_node] = elev[active_node] + thickness_at_active_node
                    #Add the neighbors to the list, but only if the thickness of the current ejecta layer was nontrivial. Smallest crater is 5m diam, so 1m deep - so its max ejecta thickness at the rim is 10cm! We should probably resolve down to, say, 2.5cm, which is <4 crater radii from the rim of our smallest crater.
                    if thickness_at_active_node > self._minimum_ejecta_thickness:
                        neighbors_active_node = grid.get_neighbor_list(active_node)
                        for x in neighbors_active_node:
                            if not flag_already_in_the_list[x]:
                                if x!=-1:
                                    crater_node_list.append(x)
                                    flag_already_in_the_list[x] = 1
        #print 'f/mu is ', f_theta/mu_theta_by_mu0
        print 'Thickness at rim: ', thickness_at_rim
        
        #Save any data to the higher level:
        self.mass_balance_in_impact = numpy.mean(elev_changes)
        self.ejecta_azimuth = _ejecta_azimuth
        self.impactor_angle_to_surface_normal = _impactor_angle_to_surface_normal

    def set_elev_change_across_grid(self, grid, data):
        '''
        This is a method to take an existing impact properties and a known nearest node to the impact site, and alter the topography to model the impact. It assumes crater radius and depth are known, models cavity shape as a power law where n is a function of R/D, and models ejecta thickness as an exponential decay,sensitive to both ballistic range from tilting and momentum transfer in impact (after Furbish). We DO NOT yet model transition to peak ring craters, or enhanced diffusion by ejecta in the strength regime. Peak ring craters are rejected from the distribution. This version of this method is designed to remove the sheer walls around the edges of craters, and replace them with a true dipping rim.
        This routine differs from set_elev_change_at_pts() as it adjusts elevations at all nodes in the grid, rather than crawling out from the central impact point to a threshold depth. This has the potential to be faster for some craters, as we can operate on the whole grid in C without looping.
        '''

        #Load in the data, for speed:
        _angle_to_horizontal=self._angle_to_horizontal
        _surface_slope=self._surface_slope
        _surface_dip_direction = self._surface_dip_direction
        _azimuth_of_travel = self._azimuth_of_travel
        pi = numpy.pi
        twopi = 2.*pi
        tan = numpy.tan
        cos = numpy.cos
        sin = numpy.sin
        sqrt = numpy.sqrt
        arctan = numpy.arctan
        where = numpy.where
        _radius = self._radius
        elev = data.elev
        
        #Derive the exponent for the crater shape, shared betw simple & complex:
        crater_bowl_exp = self.get_crater_shape_exp()
        
        #Derive the effective angle for impact, relative to the surface, beta_eff. The direction is always controlled by the impactor. Draw new impact angles if impact geomtery is impossible.
        while 1:
            #This ugly function applies a correction to the azimuth of travel which will be the actual angle the ejecta is propelled along. This is important since if, e.g., _angle_to_horizontal ~pi/2, then the surface dip direction needs to be the dominant term.
            denominator = cos(_angle_to_horizontal) + sin(_surface_slope)*cos(_surface_dip_direction - _azimuth_of_travel)
            #print 'Terms in denominator: ', cos(_angle_to_horizontal),  sin(_surface_slope), cos(_surface_dip_direction - _azimuth_of_travel)
            tan_angle_from_imp_az = sin(_surface_slope)*sin(_surface_dip_direction - _azimuth_of_travel) / denominator
            if denominator > 0.:
                _ejecta_azimuth = (_azimuth_of_travel + arctan(tan_angle_from_imp_az) + twopi)%(twopi)
            else:
                _ejecta_azimuth = (_azimuth_of_travel - arctan(tan_angle_from_imp_az) + pi)%(twopi)
            print 'Angle from impact azimuth: ', _ejecta_azimuth - _azimuth_of_travel
            beta_eff_options = [numpy.nan] * 2
            beta_eff_options[0] = _angle_to_horizontal*cos(_ejecta_azimuth - _azimuth_of_travel) - _surface_slope * cos(_ejecta_azimuth - _surface_dip_direction)
            beta_eff_options[1] = (pi - _angle_to_horizontal*cos(_ejecta_azimuth - _azimuth_of_travel)) + _surface_slope * cos(_azimuth_of_travel - _surface_dip_direction)
            beta_eff = min(beta_eff_options)
            #Need to think thru if this reversal is still required:
            #if beta_eff_options[0] > beta_eff_options[1]:
            #_azimuth_of_travel = (_azimuth_of_travel + pi)%(2.*pi)
            print 'Beta effective: ', beta_eff
            if beta_eff >= 0.:
                break
            else:
                print 'Refreshing the impactor angle'
                self.set_impactor_angles()
                _azimuth_of_travel = self._azimuth_of_travel
                _angle_to_horizontal = self._angle_to_horizontal
        _impactor_angle_to_surface_normal = 0.5*pi - beta_eff
        tan_beta = tan(_impactor_angle_to_surface_normal*self._beta_factor)
        tan_beta_sqd = tan_beta*tan_beta
        print 'Impact, ejecta azimuths: ', _azimuth_of_travel, _ejecta_azimuth
        #print 'Beta effective: ', beta_eff
        #rim_height = self._cavity_volume/(3.667*pi*_radius**2.)

        #Derive the new ejecta shape params:
        tan_repose = tan(32.*pi/180.)
        _b = 3.667*_radius*_radius*tan_repose
        _c = self._cavity_volume - 3.667*_radius*_radius*_radius*tan_repose - 0.5*_radius*_radius*tan_repose
        radius_calc = (-_b + sqrt(_b*_b - 2.*tan_repose*_c))/tan_repose
        thickness_at_rim = (_radius - radius_calc)*tan_repose
        #...where thickness(r) = thickness_at_rim*(r/R_true)**-2.75
        
        #Here the code diverges from the crawling version:
        _vec_x_to_center = grid.get_node_x_coords() - self._xcoord
        _vec_y_to_center = grid.get_node_y_coords() - self._ycoord
        _vec_r_to_center = sqrt(_vec_x_to_center*_vec_x_to_center + _vec_y_to_center*_vec_y_to_center)
        try:
            _vec_angle_to_yaxis = arctan(_vec_y_to_center/_vec_x_to_center)
        except: #These cases have the impact right on a gridline.
            _vec_theta = numpy.empty(len(_vec_x_to_center))
            nonzero_nodes = numpy.nonzero(_vec_x_to_center)
            _vec_angle_to_yaxis = arctan(_vec_y_to_center[nonzero_nodes]/_vec_x_to_center[nonzero_nodes])
            _vec_theta[nonzero_nodes] = where(_vec_x_to_center[nonzero_nodes]<0,1.5*pi-_vec_angle_to_yaxis,0.5*pi-_vec_angle_to_yaxis)
            zero_nodes = where(_vec_x_to_center==0.)[0]
            _vec_theta[zero_nodes] = where(_vec_y_to_center[zero_nodes]<0,pi,0.)
        else: #the normal case
            _vec_theta = where(_vec_x_to_center<0,1.5*pi-_vec_angle_to_yaxis,0.5*pi-_vec_angle_to_yaxis)

        #We need to account for deposition depth elevating the crater rim, i.e., we need to deposit *before* we cut the cavity. We do this by defining three domains for the node to lie in: 1. r<r_calc, i.e., below the pre-impact surface. No risk of intersecting the surface here. 2. r_calc < r; Th>z_new. this is the domain in the inward sloping rim of the crater ejecta. 3. Th<z_new and beyond. out on the ejecta proper. Note - (1) is not hard & fast rule if the surface dips. Safer is just (Th-lowering)<z_new
        #So, calc the excavation depth for all nodes, just to be on the safe side for strongly tilted geometries:
        _vec_new_z = where(_vec_r_to_center<=_radius, self.closest_node_elev - self._depth * (1. - (_vec_r_to_center/_radius)**crater_bowl_exp) + thickness_at_rim, self.closest_node_elev + thickness_at_rim + (_vec_r_to_center-_radius)*tan_repose)
    #!!Need to come back to loading data into elev_changes, which keeps track of the mass balance - tho it's less important here.
        nodes_below_surface = where(_vec_new_z<elev)[0]
        nodes_above_surface = where(_vec_new_z>=elev)[0]
        #Check if we need to adjust for a central peak
        if self._crater_type:
            central_peak_pts = where(_vec_r_to_center<=self._complex_peak_radius)
            _vec_new_z[central_peak_pts] = _vec_new_z[central_peak_pts] + self._complex_peak_str_uplift * (1. - _vec_r_to_center[central_peak_pts]/self._complex_peak_radius)
        #Set the ground elev for below ground nodes
        elev[nodes_below_surface] = _vec_new_z[nodes_below_surface]
        #From here on, our new arrays will only be as long as nodes_above_surface
        _vec_flat_thickness_above_surface = thickness_at_rim*(_vec_r_to_center[nodes_above_surface]/_radius)**-2.75
        _vec_theta_eff = _ejecta_azimuth - _vec_theta[nodes_above_surface] #This is the angle of the center-to-active-node line to the azimuth along which the ejecta is concentrated
        _vec_sin_theta_sqd = sin(_vec_theta_eff) ** 2.
        _vec_cos_theta = cos(_vec_theta_eff)
        #REMEMBER, as tan_beta gets >1, the function describing the ejecta is only valid over ever more restricted ranges of theta!! In other words,
        nodes_inside_ejecta = where(_vec_sin_theta_sqd*tan_beta_sqd <= 1.) #these are indices to an array of length nodes_above_surface only
        _vec_thickness = numpy.zeros(nodes_above_surface.size) #i.e., it's zero outside the ejecta
        _vec_mu_theta_by_mu0 = tan_beta * _vec_cos_theta[nodes_inside_ejecta] + sqrt(1. - _vec_sin_theta_sqd[nodes_inside_ejecta] * tan_beta_sqd)
        _vec_f_theta = (tan_beta_sqd*(_vec_cos_theta[nodes_inside_ejecta]**2.-_vec_sin_theta_sqd[nodes_inside_ejecta]) + 2.*tan_beta*_vec_cos_theta[nodes_inside_ejecta]*sqrt(1.-tan_beta_sqd*_vec_sin_theta_sqd[nodes_inside_ejecta]) + 1.) / twopi
        #So, distn_at_angle = distn_vertical_impact*f_theta/mu_theta_by_mu0. Draw the thickness at the active node:
        #NB-the 2pi is to correct for mismatch in the dimensions of mu and f
        thickness_at_nodes_under_ejecta = _vec_f_theta/_vec_mu_theta_by_mu0 * twopi * _vec_flat_thickness_above_surface[nodes_inside_ejecta]
        #Set the thicknesses <0 to 0:
        thickness_at_nodes_under_ejecta = where(thickness_at_nodes_under_ejecta>=0.,thickness_at_nodes_under_ejecta, 0.)
        #make the ejecta thickness map for all nodes above surface:
        _vec_thickness[nodes_inside_ejecta] = thickness_at_nodes_under_ejecta
        #Now, are we inside or outside the rim?
        elev[nodes_above_surface] = where(_vec_new_z[nodes_above_surface]<=elev[nodes_above_surface]+_vec_thickness,_vec_new_z[nodes_above_surface], elev[nodes_above_surface]+_vec_thickness)

        
        #Save any data to the higher level:
        self.mass_balance_in_impact = 0. #this isn't true, but now effectively a flag that we've used this method
        self.ejecta_azimuth = _ejecta_azimuth
        self.impactor_angle_to_surface_normal = _impactor_angle_to_surface_normal


    def excavate_a_crater(self, grid, data):
        '''
        This method executes the most of the other methods of this crater class, and makes the geomorphic changes to a mesh associated with a single bolide impact with randomized properties. It receives parameters of the model grid, and the vector data storage class. It is the primary interface method of this class.
        Unless devtesting, this method has been superceded by excavate_a_crater_optimized().
        '''
        self.set_cr_radius_from_shoemaker(data)
        #self._radius = forced_size
        print 'Radius: ', self._radius
        self.set_depth_from_size()
        self.set_crater_volume()
        self.set_coords(grid, data)
        self.set_impactor_angles()
        self.set_crater_mean_slope_v2(grid, data)
        if numpy.isnan(self._surface_slope):
            print 'Surface slope is not defined for this crater! Is it too big? Crater will not be drawn.'
        else:
            self.set_elev_change_at_pts(grid, data)
            print 'Impactor angle to ground normal: ', self.impactor_angle_to_surface_normal
            print 'Mean mass balance/px: ', self.mass_balance_in_impact
        print '*****'
        #Record the data:
        data.impact_sequence.append({'x': self._xcoord, 'y': self._ycoord, 'r': self._radius, 'volume': self._cavity_volume, 'surface_slope': self._surface_slope, 'normal_angle': self.impactor_angle_to_surface_normal, 'impact_az': self._azimuth_of_travel, 'ejecta_az': self.ejecta_azimuth, 'mass_balance': self.mass_balance_in_impact})

    def excavate_a_crater_whole_grid(self, grid, data):
        '''
            This method executes the most of the other methods of this crater class, and makes the geomorphic changes to a mesh associated with a single bolide impact with randomized properties. It receives parameters of the model grid, and the vector data storage class. It is the primary interface method of this class.
            Unless devtesting, this method has been superceded by excavate_a_crater_optimized().
            '''
        self.set_cr_radius_from_shoemaker(data)
        #self._radius = forced_size
        print 'Radius: ', self._radius
        self.set_depth_from_size()
        self.set_crater_volume()
        self.set_coords(grid, data)
        self.set_impactor_angles()
        self.set_crater_mean_slope_v2(grid, data)
        if numpy.isnan(self._surface_slope):
            print 'Surface slope is not defined for this crater! Is it too big? Crater will not be drawn.'
        else:
            self.set_elev_change_across_grid(grid, data)
            print 'Impactor angle to ground normal: ', self.impactor_angle_to_surface_normal
            print 'Mean mass balance/px: ', self.mass_balance_in_impact
        print '*****'
        #Record the data:
        data.impact_sequence.append({'x': self._xcoord, 'y': self._ycoord, 'r': self._radius, 'volume': self._cavity_volume, 'surface_slope': self._surface_slope, 'normal_angle': self.impactor_angle_to_surface_normal, 'impact_az': self._azimuth_of_travel, 'ejecta_az': self.ejecta_azimuth, 'mass_balance': self.mass_balance_in_impact})

    def excavate_a_crater_optimized(self, grid, data, forced_radius=numpy.nan, forced_angle=numpy.nan, forced_pos=numpy.nan):
        '''
            This method executes the most of the other methods of this crater class, and makes the geomorphic changes to a mesh associated with a single bolide impact with randomized properties. It receives parameters of the model grid, and the vector data storage class. It is the primary interface method of this class.
            This method is optimized to not sweep the whole grid if the crater is small.
            A fixed crater size can be specified with the input variable "forced_radius" (in km), and a fixed impact angle with "forced_angle" (in degrees from vertical - impact azimuth will always be assumed as travel eastwards). Position can be specified with forced_pos, which takes an array-like object with two entries, which are the x and y coordinate in relative position on the grid (e.g., [0.5, 0.5]).
            '''
        if numpy.isnan(forced_radius):
            self.set_cr_radius_from_shoemaker(data)
        else:
            self._radius = forced_radius
        #self._radius = forced_size
        print 'Radius: ', self._radius
        self.set_depth_from_size()
        self.set_crater_volume()
        if numpy.any(forced_pos):
            self.set_coords(grid, data)
        else:
            self._xcoord = forced_pos[0]*(mg.get_grid_xdimension()-mg.dx)
            self._ycoord = forced_pos[1]*(mg.get_grid_ydimension()-mg.dx)
            print self._xcoord, self._ycoord
        if numpy.isnan(forced_angle):
            self.set_impactor_angles()
            #print 'Angle was NaN!'
        else:
            self._angle_to_horizontal = forced_angle/numpy.pi*180.
            self._azimuth_of_travel = 0.5*numpy.pi
        self.set_crater_mean_slope_v2(grid, data)
        if numpy.isnan(self._surface_slope):
            print 'Surface slope is not defined for this crater! Is it too big? Crater will not be drawn.'
        else:
            #NB - this is an empirical optimization. If the minimum crater thickness in set_elev_change_at_pts() changes, so will the optimized value! 0.000001 min thickness == 0.024 km radius here. Normalize by R=0.0025 here => 9.6
            if self._radius/grid.dx < 9.6:
                self.set_elev_change_at_pts(grid, data)
            else:
                self.set_elev_change_across_grid(grid, data)
            print 'Impactor angle to ground normal: ', self.impactor_angle_to_surface_normal
            print 'Mean mass balance/px: ', self.mass_balance_in_impact
        print '*****'
        #Record the data:
        data.impact_sequence.append({'x': self._xcoord, 'y': self._ycoord, 'r': self._radius, 'volume': self._cavity_volume, 'surface_slope': self._surface_slope, 'normal_angle': self.impactor_angle_to_surface_normal, 'impact_az': self._azimuth_of_travel, 'ejecta_az': self.ejecta_azimuth, 'mass_balance': self.mass_balance_in_impact})


#The functions in this segment give control over the execution of this module. Adjusty the params inside the functions to get different effects, and the final line of the file to determine whether you get one crater, or lots of craters.

def dig_some_craters_on_fresh_surface():
    '''
    Ad hoc driver code to make this file run as a standalone:
    '''
    #User-defined params:
    nr = 500
    nc = 500
    dx = 0.0025
    #dt = 1.
    nt = 1000

    #Setup
    mg = RasterModelGrid()
    mg.initialize(nr, nc, dx)
    vectors = data(mg)
    vectors.elev[:] = 100.
    cr = impactor()

    #Update until
    for i in xrange(0,nt):
        print 'Crater number ', i
        cr.excavate_a_crater_optimized(mg, vectors)

    #Finalize
    elev_raster = mg.node_vector_to_raster(vectors.elev, flip_vertically=True)
    #contour(elev_raster)

    imshow(elev_raster)
    colorbar()
    show()
    vectors.viewing_raster = elev_raster
    return cr, mg, vectors

def dig_some_craters(grid, data, nt_in=10000, min_radius=0.005):
    '''
    Takes an existing DTM and peppers it with craters.
    '''
    #dt = 1.

    #Setup
    cr = impactor(min_radius)

    #Update until
    for i in xrange(0,nt_in):
        print 'Crater number ', i
        cr.excavate_a_crater_optimized(grid, data)
    
    #Finalize
    elev_raster = grid.node_vector_to_raster(data.elev, flip_vertically=True)
    #contour(elev_raster)
    
    profile = plot(elev_raster[300,:])
    xsec = plot(elev_raster[:,100])
    #imshow(elev_raster)
    #colorbar()
    #show()
    data.viewing_raster = elev_raster
    return grid, data, profile, xsec
    

def dig_one_crater(nr, nc, dx, rel_x, rel_y, radius):
    '''
    This is an ad-hoc script to dig one crater.
    '''
    #User-defined params:
    #nr = 1200
    #nc = 1200
    #dx = 0.0025
    #dt = 1.
    #nt = 1

    #Setup
    mg = RasterModelGrid()
    mg.initialize(nr, nc, dx)
    vectors = data(mg)
    vectors.elev[:] = 0.
    cr = impactor()

    cr._radius = radius
    print 'Radius: ', cr._radius
    cr.set_depth_from_size()
    print 'Depth: ', cr._depth
    cr.set_crater_volume()
    cr._xcoord = rel_x*mg.get_grid_xdimension()
    cr._ycoord = rel_y*mg.get_grid_ydimension()
    vertices_array = mg.get_nodes_around_point(cr._xcoord, cr._ycoord)
    distances_to_vertices = []
    for x in vertices_array:
        distances_to_vertices.append(numpy.sqrt((cr._xcoord-mg.node_x[x])**2. + (cr._ycoord-mg.node_y[x])**2.))
    cr.closest_node_index = vertices_array[numpy.argmin(distances_to_vertices)]
    cr.closest_node_elev = vectors.elev[cr.closest_node_index]

    cr._angle_to_horizontal = numpy.pi*0.5*2./3.
    cr._azimuth_of_travel = numpy.pi*1.5
    cr.set_crater_mean_slope_v2(mg, vectors)
    print 'Azimuth of travel: ', cr._azimuth_of_travel
    print 'Angle of ground: ', cr._surface_slope
    print 'Dip direction of ground: ', cr._surface_dip_direction
    cr.set_elev_change_across_grid(mg, vectors)
    print 'Impact angle to ground normal: ', cr.impactor_angle_to_surface_normal

    #Finalize
    elev_raster = mg.node_vector_to_raster(vectors.elev, flip_vertically=True)
    #contour(elev_raster)
    #imshow(elev_raster)
    #colorbar()
    #show()
    vectors.viewing_raster = elev_raster
    return cr, mg, vectors


def one_crater_then_degrade():
    #start_time = time.time()
    #cr, mg, vectors = dig_some_craters_on_fresh_surface()
    #cr, mg, vectors = dig_one_crater(120, 120, 0.025, 1., 0.5, 1.)
    #mg_10k, vectors_10k = dig_some_craters(mg, vectors)

    #This code builds a dictionary that contains time slices for each 10k craters hitting a surface:
    #How many times round?
    loops = 50 #500,000 craters
    #Build the dictionary:
    crater_time_sequ = {}
    profile_list = []
    xsec_list = []
    #Initialize the starting condition:
    cr, mg, vectors = dig_one_crater(1200, 1200, 0.0025, 0.5, 0.5, .75)
    #Save the starting conds:
    crater_time_sequ[0] = copy(vectors)
    #Run the loops
    for i in xrange(0,loops):
        mg, vectors, profile, xsec = dig_some_craters(mg, vectors)
        crater_time_sequ[i] = copy(vectors)
        profile_list.append(profile)
        xsec_list.append(xsec)
    show(profile_list)
    #end_time = time.time()
    #print('Elapsed time was %g seconds' % (end_time - start_time))
    return crater_time_sequ

def ten_times_reduction(mg_in, vectors_in, loops=25):
    """
    Depreciated in favour of step_reduce_size.
    """
    crater_time_sequ_50_m_min = {}
    crater_time_sequ_5_m_min = {}
    profile_list = []
    xsec_list = []
    for i in xrange(0,loops):
        mg_in, vectors_in, profile, xsec = dig_some_craters(mg_in, vectors_in, nt_in=10000, min_radius=0.05)
        crater_time_sequ_50_m_min[i] = copy(vectors_in)
    for i in xrange(0, loops):
        mg_in, vectors_in, profile, xsec = dig_some_craters(mg_in, vectors_in, nt_in=10000, min_radius=0.005)
        crater_time_sequ_5_m_min[i] = copy(vectors_in)
    return crater_time_sequ_50_m_min, crater_time_sequ_5_m_min

def step_reduce_size(mg_in, vectors_in, loops=[25,25], interval=10000, min_radius_in=[0.05, 0.005]):
    crater_time_sequ_1st = {}
    crater_time_sequ_2nd = {}
    profile_list = []
    xsec_list = []
    for i in xrange(0,loops[0]):
        mg_in, vectors_in, profile, xsec = dig_some_craters(mg_in, vectors_in, nt_in=interval, min_radius=min_radius_in[0])
        crater_time_sequ_1st[i] = copy(vectors_in)
    for i in xrange(0, loops[1]):
        mg_in, vectors_in, profile, xsec = dig_some_craters(mg_in, vectors_in, nt_in=interval, min_radius=min_radius_in[1])
        crater_time_sequ_2nd[i] = copy(vectors_in)
    return crater_time_sequ_1st, crater_time_sequ_2nd

def plot_hypsometry(plotting_rasters):
    """
    This function plots hypsometry (elev above minimum vs no. of px below it) for each of plotting rasters provided to the function. If plotting_rasters is a dictionary of data, where the plotting rasters are stored as data.plotting_raster, the function will return all entries. If it is a single plotting raster, or a single 'data' object, only that raster will plot. Call 'show()' manually after running this code.
    """
    
    def plot_a_hypsometry_curve(elevs, color=0):
        min_elev=numpy.amin(elevs)
        rel_relief = (elevs.flatten() - min_elev)/(numpy.amax(elevs)-min_elev)
        rel_relief.sort()
        hyps_x_axis = (1.-numpy.array(range(rel_relief.size), dtype=float)/rel_relief.size)
        #print hyps_x_axis
        #print rel_relief
        input_color = color*0.9
        plot(hyps_x_axis, rel_relief, color=str(input_color))
    
    if type(plotting_rasters) == dict:
        for i in range(len(plotting_rasters)):
            fraction = float(i)/len(plotting_rasters)
            plot_a_hypsometry_curve(plotting_rasters[i].viewing_raster, color=fraction)
    elif type(plotting_rasters) == numpy.ndarray:
        plot_a_hypsometry_curve(plotting_rasters)
    else:
        try:
            plot_a_hypsometry_curve(plotting_rasters.viewing_raster)
        except:
            print 'Input type not recognised!'
        
            


#if __name__=='__main__':
#    main()



#! /usr/bin/env python

from ez_setup import use_setuptools
use_setuptools()
from setuptools import setup
import multiprocessing

setup(name='TheLandlab',
      version='0.1.1',
      author='Eric Hutton',
      author_email='eric.hutton@colorado.edu',
      url='https://csdms.colorado.edu/trac/landlab',
      description='Python tools for Earth surface modeling.',
      long_description=open('README.rst').read(),
      #install_requires=['CmtBasicModelingInterface'],
      #setup_requires=['nose>=1.3'],
      setup_requires=['nose>=1.0'],
      packages=['landlab',
                'landlab.components',
                'landlab.grid',
                'landlab.field',
                'landlab.io',
                'landlab.tests',
                'landlab.utils',
                'landlab.examples',
                'landlab.framework',
               ],
      entry_points={
          'console_scripts': [
              'craters = landlab.components.craters:main',
              'landlab = landlab.cmd:main',
              'landlab_ex_1 = landlab.examples.diffusion2Dtest:main',
              'landlab_ex_2 = landlab.examples.diffusion2Dtest2:main',
              'landlab_ex_3 = landlab.examples.diffusion2Dtest3:main',
          ]
      },
      package_data={'': ['data/*asc']},
      test_suite='nose.collector',
      #test_suite='landlab.tests',
     )



######### soil_moisture_driver.py ###########
##
## This component can be considered as an example for implementation
## of soil_moisture.py
##
##
##
#############################################

# Soil Moisture Driver

from soil_moisture import *

""" Create an Soil Moisture Object """
SM = SoilMoisture()
""" Initialize Soil Moisture """
SM.initialize()
""" Update Soil Moisture """
SM.update()


""" Calculate RHS of Water Balance """ 
Total = (SM._S[SM._iterate_storm-1]-SM._si)*1000*SM._soil_pc*SM._zr+sum(SM._ETA)+sum(SM._D)+sum(SM._Ro)
""" Calculate percentage error in LHS and RHS of water balance equation """
Percent = (sum(SM._P)-Total)/(sum(SM._P))*100
print '\nPercent Error: ', Percent, '\n'


""" Plot output """
SM.plott()



from landlab.grid.base import ModelGrid
from landlab.grid.hex import HexModelGrid
from landlab.grid.radial import RadialModelGrid
from landlab.grid.raster import RasterModelGrid
from landlab.grid.voronoi import VoronoiDelaunayGrid

from landlab.grid.base import (BAD_INDEX_VALUE, INTERIOR_NODE,
                               FIXED_VALUE_BOUNDARY, FIXED_GRADIENT_BOUNDARY,
                               TRACKS_CELL_BOUNDARY, INACTIVE_BOUNDARY,
                              )
from landlab.grid.create import create_and_initialize_grid
